package cs131w.assembly;



import java.io.FileWriter;
import java.io.IOException;
import java.util.Date;
import cs131w.parser.*;
import cs131w.util.*;
import cs131w.symtab.sto.*;



public class AssemblyWriter implements Asmbl {
	
    private static int indent_level = 0;
	private static boolean secReOrdering = false;
    private static int tmpLabelIdx = 0;
    
    private static final String ERROR_IO_CLOSE = 
        "Unable to close fileWriter";
    private static final String ERROR_IO_CONSTRUCT = 
        "Unable to construct FileWriter for file %s";
    private static final String ERROR_IO_WRITE = 
        "Unable to write to fileWriter";

    private static FileWriter fileWriter;
    
    private static final String FILE_HEADER = 
        "/* DO NOT MODIFY! \n" +
        " * Auto-generated by cs131w compiler on %s\n" + 
        " * from input file '%s'\n" + 
		" */\n\n";

	private static final String AS_CONSTANTS = 
		"GLOB.1_0:	.single		0r1.0 \n" +
		"endl:		.asciz		\"\\n\" \n" +
		"intFmt:	\t.asciz		\"%d\" \n" +
		"boolFmt:	.asciz		\"%s\" \n" +
		"strFmt:	\t.asciz		\"%s\" \n" +
		"PtrT:		.asciz		\"ptr@%d\" \n" +
		"StructT:	.asciz		\"struct@%d\" \n" +
		"ArrayT:	\t.asciz		\"array@%d\" \n" +
		"boolT:		.asciz		\"true\" \n" +
		"boolF:		.asciz		\"false\" \n" +
		"ArrayIdxE:	.asciz		\"Index value of %d is outside legal range [0,%d).\\n\" \n" + 
		"NULLPtr:	.asciz		\"Attempt to dereference NULL pointer.\\n\" \n" + 
		"DeAllPtr:	.asciz		\"Attempt to dereference a pointer into deallocated stack space.\\n\" \n" +
		IND + ".section 	\".data\" \n" +
		IND + ".align		4\n" + 
		"GLOB.LowSP:\t.word		0 \n" +
		"\n";
        
	private static final String RUN_GLOB_INIT = 
		IND + "call" + SEPARATOR + "_global.INIT\n" +
		IND + "nop\n" +
		"\n";
	

	private static final String GLOBAL_INIT_PRE = 
		IND + ".section 	\".data\" \n" +
		IND + ".align		4\n" + 
		"global.INIT_ran:\n" +	
		IND + ".word\t	0\n" +
		IND + ".section 	\".text\" \n" +
		IND + ".align		4\n" + 
		"_global.INIT:		\n" +
		IND + "set 		global.INIT_ran, %l0\n" +
		IND + "ld 			[%l0], %l1\n" +
		IND + "!now compare\n" +
		IND + "cmp			%l1, %g0\n" +
		IND + "bne			after_global.INIT		! if flag already set jump back\n" +
		IND + "nop\n" + 
		IND + "set 		global.INIT.SAVE, %g1\n" + 
		IND + "save		%sp, %g1, %sp\n" +  
		IND + "set			GLOB.LowSP, %g1		!(L4) set label of global lowest SP val into G1\n" +
		IND + "ld			[%g1], %g2				!(L4) load\n" +
		IND + "cmp			%sp, %g2				!(L4) compare\n" +
		IND + "bgu			global.INIT.LowSP.end	!(L4) branch if higher\n" +
		IND + "st			%sp, [%g1]				!(L4) store current SP to global var if lower\n" +
		"global.INIT.LowSP.end:\n" + 
		"\n\n! This is where the actual code to init global vars goes !\n";

	//the actual code goes here
	
	private static final String GLOBAL_INIT_POST = 
		"\n" + IND + "set 		global.INIT_ran, %l0	! set the flag\n" +
		IND + "set	 		1, %l1					! and set flag to 1\n" +
		IND + "st	 		%l1, [%l0]				! and set flag to 1\n" +
		"after_global.INIT:								! and return back to main\n" +
		IND + "ret\n" +
		IND + "restore\n" +
		IND + "global.INIT.SAVE = ";			//add number of local bytes here
    
	//global init Stringbuilder, goes in between the GLOBAL_INIT string, where the %s is
	private static StringBuilder globalInit_SB	= new StringBuilder();
	public static int globalInit_tempSpace		= 0;
	
	private static int currentSec 	= -1;

	private static final String[] SECTION_STRINGS = { TEXT_S, DATA_S, RODATA_S, BSS_S };
	
	//global Stringbuilder
	private static StringBuilder global_SB	= new StringBuilder();

	
	//section Stringbuilders
	private static StringBuilder[] section_SBs	= {new StringBuilder(), new StringBuilder(), new StringBuilder(), new StringBuilder() };
    

	//----------------------------------------------------------------
	//
	//----------------------------------------------------------------
    public AssemblyWriter (String fileToWrite) {
        try {
            fileWriter = new FileWriter(fileToWrite);

            appendLine( String.format(FILE_HEADER, (new Date()).toString(), MyParser.getLexer().getEPFilename()) );
            switchSection(RODATA);		//from now on we assume to be in text section
			appendLine(AS_CONSTANTS);
			switchSection(TEXT);		//from now on we assume to be in text section
			flush();

		} catch (IOException e) {
            System.err.printf(ERROR_IO_CONSTRUCT, fileToWrite);
            e.printStackTrace();
            System.exit(1);
        }
    }
    
    protected void finalize() { this.dispose(); }


    public void dispose() {
        try {
			flush();			//flush all SBuffers
			
			//now write the global Init
			MyParser.debug(this + ": adding global.INIT to output!");
			
			fileWriter.write( GLOBAL_INIT_PRE );
			//see if we need to save new lowest SP
			//FuncSTO.saveLowestSP();
			fileWriter.write( globalInit_SB.toString() );
			
			fileWriter.write( GLOBAL_INIT_POST );
			String loc = new Integer(Math.abs(globalInit_tempSpace)).toString();
			
			fileWriter.write( "-(92 + "+loc+") & -8 \n\n" );
			fileWriter.close();
        } catch (IOException e) {
            System.err.println(ERROR_IO_CLOSE);
            e.printStackTrace();
            System.exit(1);
        }
    }
    
	public static void
	flush(){
		try {
			fileWriter.write(global_SB.toString());
			fileWriter.flush();
        } catch (IOException e) {
            System.err.println(ERROR_IO_CLOSE);
            e.printStackTrace();
            System.exit(1);
        }
		
		global_SB.delete(0,global_SB.length());
		
	}


	//----------------------------------------------------------------
	//		switches current section
	//----------------------------------------------------------------
	public static void
	switchSection(int newSec){
		switchSection(newSec, 4);
	}

	public static void
	switchSection(int newSec, int align){
		if (newSec != currentSec){
			currentSec = newSec;
			//and write section header to global SB
			writeLine(SECTION_STRINGS[newSec], currentSec, "switch to new section", 4);
			writeLine(String.format(ALIGN_S, "4"), currentSec, "just align everything by 4", 1);
			writeLine("\n", currentSec, "", 0);
		}
	}


	private static String
	getIndent(){
		StringBuilder sb = new StringBuilder();
		for (int i=0; i < indent_level; i++) {
            sb.append(SEPARATOR);
        }
		return sb.toString();
	}

	//----------------------------------------------------------------
	//		functions to write some text to the stringbuilders
	//----------------------------------------------------------------


	//write a line to current section
	public static void
	writeLine(String l){
		writeLine(l, currentSec, "", l.length());	
	}

	public static void
	writeLine(String l, String c){
		writeLine(l, currentSec, c, l.length());
	}


	//writes a line to a specified section
	public static void
	writeLine(String l, int sec){
		writeLine(l, sec, "", l.length());
	}

	//writes a line to a specified section
	public static void
	writeLine(String line, int sec, String comm, int parlen){
		MyParser.debug("ASW.writeLine: "+line+", sec: "+sec+", comm: "+comm+", Plen:"+parlen);
	
		//now add some tabs depending on length of param string
		int sep = 24;
		int tabs = Math.round((sep - parlen) / 4);
		int spaces = (sep - parlen);
		
		String ts = Utils.strRepeat("\t", tabs);
		String ss = Utils.strRepeat(" ", spaces);
		
		//if we are one short, just add another tab
		if ( parlen == 1) // || ((sep - parlen) % 4) == 3)
			ts += "\t";
		
		writeLineSection(line,sec,comm,ts);
		writeLineGlobal(line,sec,comm,ts);
	}

	public static void
	writeLineSection(String line, int sec, String comm, String ts){

		//MyParser.debug("ASW.writeLineSection: "+line+", sec: "+sec+", comm: "+comm);
		//first write to section SB
		//section_SBs[sec].append(getIndent());
		section_SBs[sec].append(line +  ts + "! " + getLineNum() + comm);
		
		//always write the line number
		//section_SBs[sec].append(getLineNum() + comm);
		
		section_SBs[sec].append("\n");

	}
	
	public static void
	writeLineGlobal(String line, int sec, String comm, String ts){

		
		//check if we are in right section
		if (currentSec != sec){
			switchSection(sec);
		}
		
		//now global
		//just check quickly to see if we need to add some more \t's 
		
		global_SB.append(line + ts + "! " + getLineNum() + comm);
		global_SB.append("\n");

	}
	
	public static void
	appendLine(String line){
		global_SB.append(line);
	}
	//----------------------------------------------------------------
	//	common functions to make writing of specific ops easier
	//----------------------------------------------------------------

    public static void writeSmt(String template, String ... params) {
        writeLine(String.format(template, (Object[])params), currentSec, "", getPLen(params));
    }

    public static void writeNewLine() {
        section_SBs[currentSec].append("\n");
		global_SB.append("\n");
    }
    

    public static void writeSmtAndComment(String comment, String template, String ... params) {
		writeLine(String.format(template, (Object[])params), currentSec, comment, getPLen(params));
	}

	private static int
	getPLen(String[] p){
		int i = 0;
		for (String s : p){
			i += s.length();
		}
		return i;
	}

	//----------------------------------------------------------------
	//	funcs for dealing with global init
	//----------------------------------------------------------------

	//write a line to global init
	public static void
	writeLineGlob(String l, String c, int parlen){
		
		int sep = 24;
		int tabs = Math.round((sep - parlen) / 4);
		
		String ts = Utils.strRepeat("\t", tabs);
		
		//if we are one short, just add another tab
		if ( parlen == 1) // || ((sep - parlen) % 4) == 3)
			ts += "\t";
		
		globalInit_SB.append(l + ts + "!" + getLineNum() + c);
		globalInit_SB.append("\n");
	}


    public static void writeSmtGlobInit(String template, String ... params) {
        writeLineGlob(String.format(template, (Object[])params), "", getPLen(params));
    }
    

    public static void writeSmtCommGlobInit(String comment, String template, String ... params) {
		writeLineGlob(String.format(template, (Object[])params), comment, getPLen(params));
	}

	//call immediately after the label for main()
	public static void
	addGlobalInitCall(){
		global_SB.append(RUN_GLOB_INIT);
	}

	public static int
	getGlobalFrameOffset ()
	{
		return globalInit_tempSpace;
	}
	
	//increases by spec. amount and returns old
	public static int
	increaseGlobalFrameOffset ()
	{
		return increaseGlobalFrameOffset(4);
	}
	
	public static int
	increaseGlobalFrameOffset (int by)
	{
		//int old = globalInit_tempSpace;
		globalInit_tempSpace -= by;
		return globalInit_tempSpace;
	}

	//----------------------------------------------------------------
	// funtions for creating some temp labels
	//----------------------------------------------------------------
	public static String
	getNextTempLabel(){
		return ("t_" + (++tmpLabelIdx) + "$L" + MyParser.GetLineNum());
	}

	public static String
	getNextIfTempLabel(){
		return ("If_" + (++tmpLabelIdx) + "$L" + MyParser.GetLineNum());
	}

	public static String
	getNextStrTempLabel(){
		return ("Str_" + (++tmpLabelIdx) + "$L" + MyParser.GetLineNum());
	}

	public static String
	getNextCmpTempLabel(){
		return ("Cmp_" + (++tmpLabelIdx) + "$L" + MyParser.GetLineNum());
	}

	public static String
	getNextFunTempLabel(){
		return ("_" + (++tmpLabelIdx));
	}

	//----------------------------------------------------------------
	//
	//----------------------------------------------------------------
 
    public static String getLineNum() {
        return "(L" + MyParser.getLexer().getLineNumber() + ") ";

    }


	public static void
	writeEndl(){ 
		writeSmt(SET, "endl", O0);
		writeSmt(CALL, "printf");
		writeSmt(NOP);
	}


	
}