package cs131w.assembly;



import cs131w.parser.*;



public class StorageLocation implements Asmbl {


	//where this is stored
	protected String		m_register;
	protected String		mem_base;
	protected int			mem_offset;
	
	//----------------------------------------------------------------
	// ctors
	//----------------------------------------------------------------

	public StorageLocation(){
		this(null, "", 0);
	}

	public StorageLocation(String reg){
		//this is where the address is going to live in mem
		//size of ptr = 4
		int off = MyParser.increaseFrameOffset(4);
		mem_base = FP;
		mem_offset = off;
		
		//and this is where it is right now
		m_register 	= reg;
		
		//save it for later use in its mem location, since this it's its creation time
		MyParser.asmblyWriteVarSmtComm ("Store address in "+m_register+" to " + (getAddrStr()) + " for later use", STORE, reg, getAddrStr());
		
		//now set that this is not in L7 anymore
		m_register 	= null;
		
	}
	
	public StorageLocation(String base, int off){
		this(null, base, off);
	}
	
	public StorageLocation(String reg, String base, int off){
		
		m_register 	= reg;
		mem_base	= base;
		mem_offset	= off;

	}


	//----------------------------------------------------------------
	//----------------------------------------------------------------
	public Object 
	clone() {      
		StorageLocation ret = new StorageLocation(new String(mem_base), mem_offset);
		if (isInReg()) ret.setReg(new String(m_register));
		return ret;
	}
	
	//----------------------------------------------------------------
	// funcs
	//----------------------------------------------------------------


	public String
	getReg(){
		return m_register;
	}

	public void
	setReg(String r){
		m_register = r;
	}

	public void
	clrReg(){
		m_register = null;
	}

	public boolean
	isInReg(){
		return (m_register != null);
	}
	
	public String
	getMemBase(){
		return mem_base;
	}

	public void
	setMemBase(String base){
		mem_base = base;
	}

	public int
	getMemOffset(){
		return mem_offset;
	}

	public void
	setMemOffset(int off){
		mem_offset = off;
	}

	public String
	getAddrStr(){
		String o = "";
		if (mem_offset >= 0) o = "+";
		o += mem_offset;
		
		return (mem_base + o);
	}


	public String
	toString()
	{
		if (MyParser.DEV){
			String ad = "";
			if (isInReg())
				ad = m_register;
			else
				ad = getAddrStr();
				
			return ("(RuntimeLoc)" + "@" + ad);
		}else
			return ("RuntimeLoc");

	}
	
	//loads this STO into a reg, if its in mem, doesnt care where
	public void
	AsmblyLoad(){ 
		MyParser.debug(this + ".AsmblyLoad() called!.");
		if (isInReg()) return;
		AsmblyLoad(Regs.getNextReg());
	}
	
	//loads this STO into a reg, if its in mem, do care about where it is stored into
	//also this assumes that this reg is free, does not check with Regs, or tells it to reserve
	public void
	AsmblyLoad(String into){ 
		MyParser.debug(this + ".AsmblyLoadInto("+into+") called!.");
		
		//if already loaded: return
		if (isInReg() && m_register == into) return;
		
		//if loaded but somewhere else, move it
		if (isInReg() && m_register != into){
		 	//need to move it there
			
			MyParser.asmblyWriteVarSmt(MOVE, m_register, into);
			//and free old one if it happens to be local
			Regs.FreeReg(m_register);
			m_register = into;
			return;
		}
	
		//no runtime address, check if it has base
		if (mem_base == null) return;
		
		//load into mem
		MyParser.asmblyWriteVarSmtComm ("Load '" + this + "' into " + into, LOAD, getAddrStr(), into);
		
		//now mark as in reg
		
		m_register = into;
	}


}