package cs131w.parser;

//---------------------------------------------------------------------
//
//---------------------------------------------------------------------

import java.util.Deque;
import java.util.Vector;
import java.util.concurrent.LinkedBlockingDeque;

import java_cup.runtime.*;
import cs131w.assembly.*;
import cs131w.error.ErrorMsg;
import cs131w.error.ErrorPrinter;
import cs131w.error.Formatter;
import cs131w.lexer.Lexer;
import cs131w.lexer.Token;
import cs131w.stmts.IfStmt;
import cs131w.stmts.WhileStmt;
import cs131w.symtab.Scope;
import cs131w.symtab.SymbolTable;
import cs131w.symtab.ops.Operator;
import cs131w.symtab.ops.binary.AndOp;
import cs131w.symtab.ops.binary.AssignOp;
import cs131w.symtab.ops.binary.OrOp;
import cs131w.symtab.sto.ConstSTO;
import cs131w.symtab.sto.ErrorSTO;
import cs131w.symtab.sto.FuncSTO;
import cs131w.symtab.sto.ParamSTO;
import cs131w.symtab.sto.STO;
import cs131w.symtab.sto.TypedefSTO;
import cs131w.symtab.sto.UeberFuncSTO;
import cs131w.symtab.sto.VarSTO;
import cs131w.symtab.types.ArrayType;
import cs131w.symtab.types.ErrorType;
import cs131w.symtab.types.FuncProtoType;
import cs131w.symtab.types.FuncPtrType;
import cs131w.symtab.types.IntType;
import cs131w.symtab.types.StructType;
import cs131w.symtab.types.Type;



public class MyParser extends parser implements Asmbl
{
	
	//Class Variable denoting if we are in dev mode, 
	//!!!!!!! set to FALSE when shipping !!!!!!!!!! 
	public static boolean DEV	= false;
	//set to true for Project Part II, will supress any output
	public static boolean QUIET	= false;
	

		public static void debug(String message){
			if(DEV){
				System.out.println("("+m_lexer.getLineNumber()+") " + message);
			}
		}


		private static Lexer			m_lexer;
		private static SymbolTable		m_symtab;
		private static AssemblyWriter	m_Asw;
		private static final String		m_AssemblyFile = "rc.s";
		
		
		private static ErrorPrinter		m_errors;
		private static int 				m_nNumErrors;
		private static int				m_nSavedLineNum;

		//the line that the last error was printed for
		//supresses duplicate errors for any single line
		private static int				m_lastErrorLine = -1;
		public static boolean			m_allowDupErrorLine = false;

		//----------------------------------------------------------------
		//	Instance variables
		//----------------------------------------------------------------
		private String					m_strLastLexeme;
		private boolean					m_bSyntaxError = true;
		private int						m_inWhile = 0;
		
		private Deque<IfStmt>			m_ifStack = new LinkedBlockingDeque<IfStmt>();
		private Deque<WhileStmt>		m_loopStack = new LinkedBlockingDeque<WhileStmt>();
		private Deque<OrOp>				m_orStack =  new LinkedBlockingDeque<OrOp>();
		private Deque<AndOp>			m_andStack = new LinkedBlockingDeque<AndOp>();
		

	//----------------------------------------------------------------
	// ctors
	//----------------------------------------------------------------
	public 
	MyParser (Lexer lexer, ErrorPrinter errors)
	{
		m_lexer = lexer;
		m_symtab = new SymbolTable ();
		m_errors = errors;
		m_nNumErrors = 0;
		m_Asw = new AssemblyWriter(m_AssemblyFile);
		//write out first line comment
		m_lexer.dumpLineComment();
		
	}


	//----------------------------------------------------------------
	// some useful functions
	//----------------------------------------------------------------
	public boolean
	Ok ()
	{
		return (m_nNumErrors == 0);
	}


	//----------------------------------------------------------------
	//
	//----------------------------------------------------------------
	public Symbol
	scan ()
	{
		Token		t = m_lexer.GetToken ();

		//	We'll save the last token read for error messages.
		//	Sometimes, the token is lost reading for the next
		//	token which can be null.
		m_strLastLexeme = t.GetLexeme ();

		switch (t.GetCode ())
		{
			case sym.T_ID:
			case sym.T_ID_U:
			case sym.T_STR_LITERAL:
			case sym.T_FLOAT_LITERAL:
			case sym.T_INT_LITERAL:
			case sym.T_CHAR_LITERAL:
				return (new Symbol (t.GetCode (), t.GetLexeme ()));
			default:
				return (new Symbol (t.GetCode ()));
		}
	}


	//----------------------------------------------------------------
	//
	//----------------------------------------------------------------
	public void
	syntax_error (Symbol s)
	{
	}


	//----------------------------------------------------------------
	//
	//----------------------------------------------------------------
	public void
	report_fatal_error (Symbol s)
	{
		m_nNumErrors++;
		if (m_bSyntaxError)
		{
			m_nNumErrors++;

			//	It is possible that the error was detected
			//	at the end of a line - in which case, s will
			//	be null.  Instead, we saved the last token
			//	read in to give a more meaningful error 
			//	message.
			m_errors.print (Formatter.toString (ErrorMsg.syntax_error, m_strLastLexeme));
		}
	}


	//----------------------------------------------------------------
	//
	//----------------------------------------------------------------
	public void
	unrecovered_syntax_error (Symbol s)
	{
		report_fatal_error (s);
	}


	//----------------------------------------------------------------
	//
	//----------------------------------------------------------------
	public void
	DisableSyntaxError ()
	{
		m_bSyntaxError = false;
	}

	public void
	EnableSyntaxError ()
	{
		m_bSyntaxError = true;
	}


	//----------------------------------------------------------------
	//
	//----------------------------------------------------------------
	public static String 
	GetFile ()
	{
		return (m_lexer.getEPFilename ());
	}

	public static int
	GetLineNum ()
	{
		return (m_lexer.getLineNumber ());
	}

	public static void
	SaveLineNum ()
	{
		m_nSavedLineNum = m_lexer.getLineNumber ();
	}

	public static int
	GetSavedLineNum ()
	{
		return (m_nSavedLineNum);
	}

	//----------------------------------------------------------------
	//
	//----------------------------------------------------------------
	public static Lexer
	getLexer ()
	{
		return (m_lexer);
	}

	public static SymbolTable
	getSymTab ()
	{
		return (m_symtab);
	}

	public static boolean
	isInGlobalSpace ()
	{
		return (m_symtab.getLevel() == 1);
	}

	public static int
	increaseFrameOffset ()
	{
		return MyParser.increaseFrameOffset(4);
	}
	
	public static int
	increaseFrameOffset (int by)
	{
		if (isInGlobalSpace())
			return AssemblyWriter.increaseGlobalFrameOffset(by);
		else if (m_symtab.isDefiningFunc() && m_symtab.getFunc().isFunc())
			return ((FuncSTO)m_symtab.getFunc()).increaseFrameOffset(by);
		
		return 0;
	}

	public static int
	getFrameOffset ()
	{
		if (isInGlobalSpace())
			return AssemblyWriter.getGlobalFrameOffset();
		else if (m_symtab.isDefiningFunc() && m_symtab.getFunc().isFunc())
			return ((FuncSTO)m_symtab.getFunc()).getFrameOffset();
		
		return 0;
	}

	public static void
	asmblyWriteVarSmt (String t, String ... p){
		MyParser.asmblyWriteVarSmtComm("", t, p);
	}
	public static void
	asmblyWriteVarSmtComm (String c, String t, String ... p)
	{
		if (isInGlobalSpace()){
			AssemblyWriter.writeSmtCommGlobInit(c,t,(String[])p);
		}else{ 	//if (m_symtab.isDefiningFunc() && m_symtab.getFunc().isFunc())
			AssemblyWriter.writeSmtAndComment(c,t,(String[])p);
		}	
	}


	//----------------------------------------------------------------
	//	private functions to make the repetetetive task of error 
	//	reporting a little easier
	//----------------------------------------------------------------
	
	//these 2 funcs take an error msg as String and actually print it to screen
	//remember last line that an error was printed for and does not print if duplicate
	private static void E_doPrintError(String e){ E_doPrintError(e,false); }

	private static void E_doPrintError(String e, boolean offset){
		int newErrorLine = m_lexer.getLineNumber();
		m_nNumErrors++;
		int lastErrorLine = m_lexer.getLineNumber();
		if (offset){
			lastErrorLine = (GetSavedLineNum());
		}
		debug("E_doPrintErr at new line " + newErrorLine + ", lastline: " + lastErrorLine + ", offs: " + offset + ", m_last: " + m_lastErrorLine);
		
		if (m_allowDupErrorLine || (m_lastErrorLine != lastErrorLine)){
			int off = 0;
			//see if we have to pass an offset, meaning we have an error that actually belongs on a line before this
			if (offset){
				off = (GetSavedLineNum() - GetLineNum());
				newErrorLine = GetSavedLineNum();
			}
			debug("printing error at line "+newErrorLine+", lastline: "+m_lastErrorLine+", file: "+m_lexer.getEPFilename ());
			
			if (!QUIET)
				m_errors.print (e, off);
			
			//now do an output in the asmbly as well, just to have an indication that something went wrong
			asmblyWriteVarSmt(ERROR, e);
			
			
			m_lastErrorLine = newErrorLine;
		}else{
			debug("Not printing error at line "+newErrorLine+", lastline: "+m_lastErrorLine+", file: "+m_lexer.getEPFilename ());
		}
	}

	//following 2 funcs return a Type (ErrorType in this case)
	// ---------------------------------------------------------
	private ErrorType
	Type_doErrorReport(String e){ return Type_doErrorReport(e,false); }

	private ErrorType
	Type_doErrorReport(String e, boolean o){
		E_doPrintError (e,o);
		return (new ErrorType());
	}

	//all these return a STO (ErrorSTO)
	// ---------------------------------------------------------
	public static ErrorSTO
	STO_doErrorReport(String e){ return STO_doErrorReport(e,false); }
	
	public static ErrorSTO
	STO_doErrorReport(String e, boolean o){
		return STO_doErrorReport(null,e,o);
	}

	//these take any object s as source of error, and try to make a sensible
	//ErrorSTO out of it
	// ---------------------------------------------------------
	public static ErrorSTO
	STO_doErrorReport(Object s, String e){ return STO_doErrorReport(s,e,false); }

	public static ErrorSTO
	STO_doErrorReport(Object s, String e, boolean o){
		E_doPrintError (e,o);
		if (s == null) return new ErrorSTO();
		else if(s instanceof ErrorSTO) return ((ErrorSTO)s);
		else return (new ErrorSTO(s));
	}


	//this prints the error for a redeclared ID, switches between appr. Msg depending
	//on if we are defining a struct
	// ---------------------------------------------------------
	private ErrorSTO
	STO_doRedeclIDReport(String id){ return STO_doRedeclIDReport(id,false); }
	
	private ErrorSTO
	STO_doRedeclIDReport(String id, boolean o){
		if (!m_symtab.isDefiningStruct())
			return STO_doErrorReport (id, Formatter.toString(ErrorMsg.redeclared_id, id), o);
		else
			return STO_doErrorReport (id, Formatter.toString(ErrorMsg.error13a_Struct, id), o);
	}

	public static ErrorSTO
	STO_getRedeclIDReport (String id){
		if (!m_symtab.isDefiningStruct())
			return new ErrorSTO (Formatter.toString(ErrorMsg.redeclared_id, id) );
		else
			return new ErrorSTO (Formatter.toString(ErrorMsg.error13a_Struct, id) );
	}

	//and this does the same with a STO that's the source of the error
	// ---------------------------------------------------------
	private ErrorSTO
	STO_doRedeclIDReport(STO sto, String id){ return STO_doRedeclIDReport(sto,id,false); }

	private ErrorSTO
	STO_doRedeclIDReport(STO sto, String id, boolean o){
		if (!m_symtab.isDefiningStruct())
			return STO_doErrorReport (sto, Formatter.toString(ErrorMsg.redeclared_id, id), o);
		else
			return STO_doErrorReport (sto, Formatter.toString(ErrorMsg.error13a_Struct, id), o);
	}

	//----------------------------------------------------------------
	//
	//----------------------------------------------------------------
	void
	DoProgramStart()
	{
		// Opens the global scope.
		m_symtab.openScope ();
		//debug("asw is " + m_Asw);
		
	}


	//----------------------------------------------------------------
	//
	//----------------------------------------------------------------
	void
	DoProgramEnd()
	{
		//debug("calling dispose on " + m_Asw);
		m_symtab.closeScope ();
		m_Asw.flush();
		m_Asw.dispose();
	}



	//----------------------------------------------------------------
	//			Declarations parsing 
	//----------------------------------------------------------------

	//----------------------------------------------------------------
	//			Declarations parsing 
	//----------------------------------------------------------------
	//----------------------------------------------------------------
	//	Vector of STOs, or a name and vector of 
	//----------------------------------------------------------------
	void
	DoVarDecl (Vector v, Type t)
	{
		this.DoVarDecl(v, t, true, false);
	}
	
	void
	DoVarDecl (Vector lstIDs, Type t, boolean strVector, boolean bStatic)
	{
		debug("var decl w/: "+t+", list: " + lstIDs);
		
		if (t!=null && t.isError())
			return;
			
		String id = null;
		STO sto = null;
		
		for (int i = 0; i < lstIDs.size (); i++)
		{
			
			if (strVector){
				//this means its a vector of strings, 
				id = (String) lstIDs.elementAt (i);
			}else{
				//and this is a vector of STOs that have the name and the expr trying to be assigned here
				sto = (STO) lstIDs.elementAt (i);
				id = sto.getName();
			}
			
			//use the varSTO's own factory method
			STO var = VarSTO.create (id, t);
			
			
			if (m_symtab.isDefiningStruct()){
				//just remember for later, that we are a member of a struct
				var.setIsStructMember(m_symtab.getStruct());
			}
			
			//if its an error, just print it and continue
			if (var.isError()){
				STO_doErrorReport (var, var.getName());
				continue;
			} 
			
			Operator o = new AssignOp();
			STO res = null;
			//now if we are assigning, just use the AssignOP to do the checking for us
			if (!strVector && !sto.getParentType().isJoker()){
				
				res = o.checkOperands(var, sto);
				if (res.isError()) {
					STO_doErrorReport (sto, Formatter.toString(ErrorMsg.error8_Assign, sto.getType(), t));
					continue;
				}
			}
			
			if (!var.isError()){
				m_symtab.insert (var);
				debug("Adding var '"+var+"' with type "+t+" to " + m_symtab);
	
				if (bStatic)
					var.setStatic();
				
				//debug("mParser.varDecl calling asmblyCreate with sto=" + sto);
				var.AsmblyCreate(sto);
				//debug("mParser.varDecl 'asmblyCreate' returned!");
				
				//now if its assigned but its not assgined with something const, do the assign op
				if (!strVector && sto != null && !sto.getParentType().isJoker() && !sto.isConst()){
					//debug("mParser.varDecl calling asmblyOperate with res=" + res + ", trg=" + var + ", src=" + sto);
					//do the guard against static init > once, only gets printed if actually static
					((VarSTO)var).AsmblyStartStaticInitGuard();
					res = o.AsmblyOperate(res, var, sto);	
					((VarSTO)var).AsmblyFinishStaticInitGuard();
				}
			}
		
		}//end for
	}


	//----------------------------------------------------------------
	//
	//----------------------------------------------------------------
	void
	DoExternDecl (Vector lstIDs, Type t)
	{
		for (int i = 0; i < lstIDs.size (); i++)
		{
			String id = (String) lstIDs.elementAt (i);

			if (m_symtab.accessLocal (id) != null)
			{
				STO_doRedeclIDReport (id);
				return;
			}

			t.setIsIncomplete(true);
			
			VarSTO 		sto = new VarSTO (id, t);
			
			//mark it as extern
			sto.markExtern();
			
			m_symtab.insert (sto);
			debug("Adding extern var '"+id+"' to " + m_symtab);
		}
	}


	//----------------------------------------------------------------
	//
	//----------------------------------------------------------------
	void
	DoConstDecl (Vector lstIDs, Type t)
	{
		DoConstDecl(lstIDs, t, false);
	}

	void
	DoConstDecl (Vector lstIDs, Type t, boolean bStatic)
	{
		
		debug("const decl w/: "+t+", list: " + lstIDs);
		
		//here we already check if its an error, or null
		if (t!=null && t.isError())
			return;
		
		for (int i = 0; i < lstIDs.size (); i++)
		{
			STO element = (STO) lstIDs.elementAt (i);
			
			//now just let ConstSTO create itself for us
			STO var = ConstSTO.create (element, t);
			
			//if its an error, just print it and continue
			if (var.isError()){
				STO_doErrorReport (var, var.getName());
				continue;
			}
			
			MyParser.debug("Adding const '"+var+"' with type "+t+" to " + m_symtab);
			m_symtab.insert (var);
			
			if (m_symtab.isDefiningStruct()){
				//just remember for later, that we are a member of a struct
				var.setIsStructMember(m_symtab.getStruct());
			}
			
			if (bStatic)
				var.setStatic();
			var.AsmblyCreate();
			
			
		}//end for
	}


	//----------------------------------------------------------------
	//
	//----------------------------------------------------------------
	void
	DoTypedefDecl (Vector lstIDs, Type t)
	{
		if (t!=null && t.isError())
			return;
		
		for (int i = 0; i < lstIDs.size (); i++)
		{
			String id = (String) lstIDs.elementAt (i);

			if (m_symtab.accessLocal (id) != null)
			{
				STO_doRedeclIDReport (id);
				return;
			}
		
			TypedefSTO 	sto = new TypedefSTO (id, t);
			m_symtab.insert (sto);
			debug("Adding type '"+id+"' with basetype '" + t + "' to " + m_symtab);
			
		}
	}


	//----------------------------------------------------------------
	// struct functions
	//----------------------------------------------------------------
	//this starts a new struct def.
	void
	DoStructdefDecl_1 (String id)
	{

		StructType struct = new StructType(id);
		struct.setIsIncomplete(true);
		
		STO sto = new TypedefSTO (id, struct);

		if (m_symtab.accessLocal (id) != null)
		{
			sto = STO_doRedeclIDReport (sto, id);
		}
		
		debug("Adding struct '"+id+"' to" + m_symtab);

		m_symtab.insert (sto);
		m_symtab.setStruct (sto);
		Scope s_scope = m_symtab.openScope ();
		struct.setScope(s_scope);
		
		//return struct;
		
	}

	//this ends a struct def and closes its scope
	void
	DoStructdefDecl_2 ()
	{
	
		STO s = m_symtab.getStruct();
		
		if (s == null)
		{
			if (DEV)
				E_doPrintError ("internal: DoStructdefDecl_2 says no struct!");
			return;
		}

		if (s.isError())
			s = ((ErrorSTO)s).getBase();

		TypedefSTO struct = (TypedefSTO)s;
	
		struct.getParentType().setIsIncomplete(false);

		m_symtab.closeScope ();
		m_symtab.setStruct (null);
		
	}


	//----------------------------------------------------------------
	// does a func declaration, type is the func prototype
	//----------------------------------------------------------------
	STO
	DoFuncDecl_1 (String id, Type t)
	{
		debug("do funcDecl_1 for '" + id + "' type " + t);
		
		
		STO var = UeberFuncSTO.start_create(id,t);
		
		debug("do funcDecl_1 for has created func: '" + var);
		
		//if its an error, just print it and continue
		if (var.isError()){
			STO_doErrorReport (var, var.getName());
		}else if (var.isUeberFunc()){
			m_symtab.insert (var);
		}
		Scope s = m_symtab.openScope ();
		((FuncSTO)m_symtab.getFunc()).setScope(s);
		
		return var;

	}



	//----------------------------------------------------------------
	//	//end a function declaration
	//----------------------------------------------------------------
	void
	DoFuncDecl_2 ()
	{
		
		STO var = UeberFuncSTO.finish_parsing();
		
		m_symtab.closeScope ();
		m_symtab.setFunc (null);
		
		//if its an error, just print it and continue
		if (var.isError()){
			m_allowDupErrorLine = true;
			STO_doErrorReport (var, var.getName(), true);
			m_allowDupErrorLine = false;
		}
	}


	
	//returns a FunctionProtoType, with the given type as the return type
	FuncProtoType
	DoFunctionProto(Vector params, Type ret){
		return FuncProtoType.create(params,ret);
	}
	
	
	void
	insertParamsIntoFuncST(Vector params){
		// insert parameters
		for (int i = 0; i < params.size (); i++)
		{
			ParamSTO var = (ParamSTO) (params.elementAt(i));
			//add into scope which should be currently scoped to function we are defining
			//or we wouldnt be able to use the params inside the function's body
				m_symtab.insert (var);
			debug("adding var '"+var+"' to local func symtab: " + m_symtab);
		}
	}
	
	
	void
	DoReturnStmt (STO ret){
		STO func = m_symtab.getFunc();
		
		if (func == null)
		{
			if (DEV)
				E_doPrintError ("internal: DoReturnStmt says no proc!");
			return;
		}

		if (func.isError())
			func = ((ErrorSTO)func).getBase();

		STO var = ((FuncSTO)func).DoReturnStmt(ret);
		
		if (var.isError())
			STO_doErrorReport (var, var.getName());
		
		else
			//now output asmbly code
			((FuncSTO)func).AsmblyReturn(ret);
		

	}


	//----------------------------------------------------------------
	//
	//----------------------------------------------------------------
	void
	DoBlockOpen ()
	{
		// Open a scope.
		m_symtab.openScope ();
	}


	//----------------------------------------------------------------
	//
	//----------------------------------------------------------------
	void
	DoBlockClose()
	{
		m_symtab.closeScope ();
	}



	//----------------------------------------------------------------
	//			Expression parsing 
	//----------------------------------------------------------------
	//----------------------------------------------------------------
	//
	//----------------------------------------------------------------


	//----------------------------------------------------------------
	//
	//----------------------------------------------------------------
	public static STO 
	DoBinaryExpr(STO a, Operator o, STO b) {
		//check if one already has an error, in that case just return
		if (a==null || a.isError() || o==null || b==null) return a;
		if (b.isError()) return b;
		STO sto;
		
		//this checks if the operands are of the approp. type and returns an error if necc.
		sto = o.checkOperands(a, b);
		if (sto.isError()) {
			return STO_doErrorReport (sto, sto.getName());
		}
		
		//constant folding is now done in operate, which also returns the appr. type
		//do explicit cast, if const, else the appropriate overriden method in subOp will not be called
		try{

			if (a.isConst() && b.isConst()){
				sto = o.operate((ConstSTO) a, (ConstSTO) b);
				//dont do the op if both ops are const.. this just gets done in java directly
			}else{
				sto = o.operate(a,b);
				//now we can do the binary op in Assembly
				sto = o.AsmblyOperate(sto, a, b);
			}
			
			
		}catch(ArithmeticException e){
			return STO_doErrorReport(ErrorMsg.error8_Arithmetic);
		}
		
		
		return sto;
		
	}

	public static STO 
	DoUnaryExpr(Operator o, STO a) {
		debug("Doing unary operator " + o + " on " + a+ ", isModifiable: " + a.isModLValue());
		if (a==null || a.isError()) return a;
		if (o==null) return a;

		STO sto = o.checkOperand(a);
		if (sto.isError()) {
			return STO_doErrorReport (sto, sto.getName());
		}
		
		if (a.isConst())
			sto = o.operate((ConstSTO)a);
		else
			sto = o.operate(a);
			o.AsmblyOperate(sto, a);
		
		return sto;
		
	}


	//----------------------------------------------------------------
	//
	//----------------------------------------------------------------

	void
	DoCondStmt (STO a){
		if (a.isError()) return;
		if (!a.getType().isBool()){
			STO_doErrorReport (a, Formatter.toString(ErrorMsg.error4_Test, a.getType()));
		}
	}

	void
	EnterWhile (){
		m_inWhile++;
	}


	void
	LeaveWhile (){
		m_inWhile--;
	}


	void
	CheckBreak(){
		if (m_inWhile <= 0){
			STO_doErrorReport (ErrorMsg.error12_Break);
		}
	}
	
	void
	CheckContinue(){
		if (m_inWhile <= 0){
			STO_doErrorReport (ErrorMsg.error12_Continue);
		}
	}
	
	//----------------------------------------------------------------
	// parses a function call to UeberFuncSTO given in sto with params p,
	// which is a vector of types STO 
	//----------------------------------------------------------------
	STO
	DoOverloadedFuncCall (STO sto, Vector p)
	{
		if ( sto.isError() || (!sto.isFunc() && !sto.isFuncPtr() && !sto.isUeberFunc()) )
		{
			return STO_doErrorReport (sto, Formatter.toString(ErrorMsg.not_function, sto.getName()));
		}
		
		if (sto.isFuncPtr())
			return this.DoFuncCall(sto, p);
		
		STO var = ((UeberFuncSTO) sto).DoFuncCall(p);
		
		if (var.isError()){
			return STO_doErrorReport (var, var.getName());
		} 
		
		debug ("overloaded-funcall doing actual funccall to: " + var + ", with: " + p);
		//now do the funccall
		return this.DoFuncCall(var, p);

	}
	
	//----------------------------------------------------------------
	// parses a function call to FuncSTO given in sto with params p,
	// which is a vector of types STO 
	//----------------------------------------------------------------
	STO
	DoFuncCall (STO sto, Vector p)
	{
		debug ("funcall calling: " + sto + " with "+p);
		
		if ( (!sto.isFunc() && !sto.isFuncPtr()) || sto.isError())
		{
			return STO_doErrorReport (sto, Formatter.toString(ErrorMsg.not_function, sto.getName()));
		}

		//now get the funcProto out of the underlying type
		FuncProtoType f;

		if (sto.isFunc()){
			f = (FuncProtoType) ((FuncSTO)sto).getProtoType();
		}else{	//FuncPtr
			f = (FuncProtoType) (sto.getType().getRootType().getPointedToType());
			 
		}
		
		STO res = f.DoFuncCall(p);
		
		
		//now do the call in asmbly
		if (sto.isFunc()){
			//normal function
			((FuncSTO)sto).AsmblyCall(p, res);
		
		}else{	//FuncPtr
			//load ptr
			sto.AsmblyLoad();
			
			String strThisPtr = null;
			
			//check if its part of a struct
			if (sto.hasStructThisPtr()){
				//yes, load that into O0
				sto.mem_StructThisPtr.AsmblyLoad(O0);
				strThisPtr = O0;
			}
			
			//call
			f.AsmblyCall(p, res, strThisPtr, sto.getReg());
			
			//and clear from reg
			sto.AsmblyClearReg();
			
		}
		
		//and store the res back
		res.AsmblyStore();
		//and return
		return res;
	}



	//----------------------------------------------------------------
	//
	//----------------------------------------------------------------
	STO
	DoThis ()
	{
		if (!m_symtab.isDefiningStruct()){
			if (DEV)
				E_doPrintError ("internal: DoThis says no struct!");
			return new ErrorSTO("this");
		}
		
		
		STO res = m_symtab.getStruct();
		debug ("Doing THIS for: " + res);
		
		//in assemlby we just need to load from %fp+68 into something
		//so let the res know that it's runtime address is at that location
		res.setRuntimeAddr(new StorageLocation(FP, 68));
		
		//just return the current struct as a TypedefSTO
		return res;
		
	}

	
	//----------------------------------------------------------------
	//	false == new, true == delete
	//----------------------------------------------------------------
	void
	DoNewDelete(STO sto, boolean newDel){
		if (!newDel){
			debug("Calling new on " + sto + ". size=" + sto.getParentType().getRootType().getPointedToType().getSize());
			debug("parentType: " + sto.getParentType().getRootType() + ". PointedToT:" + sto.getParentType().getRootType().getPointedToType());
		
		}	
		
		if (sto == null || sto.isError()) return;

		//check that it is a modifiable LVal
		if (!sto.isModLValue()){
				STO_doErrorReport(sto, ( (newDel) ? ErrorMsg.error16_Delete_var : ErrorMsg.error16_New_var));
				return;
		}

		//we have to check that this is indeed a pointer
		if (!sto.getParentType().isBasicPtr()){
				STO_doErrorReport(sto, Formatter.toString(( (newDel) ? ErrorMsg.error16_Delete : ErrorMsg.error16_New), sto.getParentType()));
				return;
		}
		
		//load ptr up
		
		//now do the op in Asmbly, false = new
		if (!newDel){
			sto.AsmblyLoad();
			
			AssemblyWriter.writeSmtAndComment("NumElem=1", SET, "1", O0);
			//get size of underlying element
			int size = sto.getParentType().getRootType().getPointedToType().getSize();
			AssemblyWriter.writeSmtAndComment("sizeof:" + sto, SET, new Integer(size).toString(), O1);
			AssemblyWriter.writeSmtAndComment("call calloc", CALL, "calloc");
			AssemblyWriter.writeSmtAndComment("wait for it... wait ... now go!", NOP);
			
			//tell the ptr that it is in O0
			sto.setReg(O0);
			//and store back
			sto.AsmblyStore();
			
		}else{
			//delete
			
			sto.AsmblyLoad(O0);
			
			AssemblyWriter.writeSmtAndComment("call free", CALL, "free");
			AssemblyWriter.writeSmtAndComment("wait for it... wait ... now GO!", NOP);
			//set to 0
			AssemblyWriter.writeSmtAndComment("set ptr to NULL", SET, NULL, L0);
			sto.setReg(L0);
			//and store
			sto.AsmblyStore();
			
			
		}
		

	}	

	//----------------------------------------------------------------
	//
	//----------------------------------------------------------------
	Type 
	DoArrayCreate(STO sto){
		
		
		Type ret = ArrayType.create (sto);
		
		//if its an error, just print it and continue
		if (ret.isError()){
			STO_doErrorReport (ret, ret.getName());
		}
		return ret;
		
	}


	//----------------------------------------------------------------
	// retain old call for backward comp.
	//----------------------------------------------------------------


	//does a designator check for any var, also an overloaded function along the way
	STO
	DoDesignator3_ID (String strID)
	{

		STO		sto;
		
		int 	skip = 0;
		
		debug ("doDes3_ID searching: " + strID + ", in " + m_symtab);
		
		//if we are inside defining a struct, skip local function and struct scope (has to accessed with 'this')
		if (m_symtab.isDefiningStruct()){
			//instead we check the local scope first, and then the global one in order
			if ( 
					((sto = m_symtab.accessLocal (strID)) == null)
				&&
					((sto = m_symtab.accessGlobal (strID)) == null)
				)
			{
				return STO_doErrorReport (strID, Formatter.toString(ErrorMsg.undeclared_id, strID));
			}	
			
		}else{
			debug ("doDes3_ID symtab.access(" + strID + "): " + m_symtab.access(strID));
			if ((sto = m_symtab.access (strID)) == null)
			{
			 	return STO_doErrorReport (strID, Formatter.toString(ErrorMsg.undeclared_id, strID));	
			}
		}
		
		
		//return constSTO if it is a name of a function, but not an ueberfunc
		debug("doing qual3 for id: "+strID + ",sto is:"+sto);
		if (sto.isFunc() && !sto.isUeberFunc()){
			//generate new const function pointer ConstSTO (String strName, Type t, Double val)
			sto = new ConstSTO(strID, new FuncPtrType(((FuncSTO)sto).getProtoType()), new Double(0));
		}
		
		//_asmbly:
		//load this puppy into a reg since we are likely to use it for something
		//sto.AsmblyLoad();
		
		debug ("doDes3_ID returning: " + sto);
		return (sto);
	}


	//----------------------------------------------------------------
	//
	//----------------------------------------------------------------
	STO
	DoQualIdent (String strID)
	{
		STO		sto;

		if ((sto = m_symtab.access (strID)) == null)
		{
		 	return STO_doErrorReport (strID, Formatter.toString(ErrorMsg.undeclared_id, strID));	
		}

		if (!sto.isTypedef())
		{
			return STO_doErrorReport (sto, Formatter.toString(ErrorMsg.not_type, sto.getName ()));
		}

		return (sto);
	}
	
	
	//----------------------------------------------------------------
	//
	//----------------------------------------------------------------
	void
	DoExitStmt(STO sto){
		if (!sto.getType().isAssignable(new IntType())){
			STO_doErrorReport (sto, Formatter.toString(ErrorMsg.error7_Exit, sto.getType()));
		}
		
		/* this is what cc -S produces for exit stmt, is this right?
				call    exit
			    mov     %g0,%o0
			    jmp     %i7+8		 -> this is ret?
			    restore
		
			do the same:
		*/
		
		AssemblyWriter.writeSmt(CALL, "exit");
		sto.AsmblyLoad(O0);
		
		AssemblyWriter.writeSmt(RET);
		AssemblyWriter.writeSmt(RESTORE);
		
		
	}
	
	public void AsmblyDoBreak(){
		asmblyWriteVarSmtComm("break", BA, peekLoop().getEndLabel());
		asmblyWriteVarSmtComm("break", NOP);
	}
	
	public void AsmblyDoContinue(){
		asmblyWriteVarSmtComm("continue", BA , peekLoop().getTopLabel());
		asmblyWriteVarSmtComm("continue", NOP);
	}
	
	public void doStmt(){
		//Regs.printRegStatusComment();
	}
	
	public void pushIf(IfStmt stmt){
		m_ifStack.push(stmt);
	}
	
	public IfStmt popIf(){
		return (IfStmt) m_ifStack.pop();
	}
	
	public IfStmt peekIf(){
		return (IfStmt) m_ifStack.peek();
	}
	
	public void pushLoop(WhileStmt stmt){
		m_loopStack.push(stmt);
	}
	
	public WhileStmt peekLoop(){
		return m_loopStack.peek();
	}
	
	public WhileStmt popLoop(){
		return m_loopStack.pop();
	}
	
	public void pushOr(OrOp op){
		m_orStack.push(op);
	}
	
	public OrOp peekOr(){
		return m_orStack.peek();
	}
	
	public OrOp popOr(){
		return m_orStack.pop();
	}	
	
	public void pushAnd(AndOp op){
		m_andStack.push(op);
	}
	
	public AndOp peekAnd(){
		return m_andStack.peek();
	}
	
	public AndOp popAnd(){
		return m_andStack.pop();
	}
	
}
