package cs131w.symtab;
//---------------------------------------------------------------------
//
//---------------------------------------------------------------------

import java.util.Vector;
import cs131w.symtab.sto.*;
import cs131w.symtab.types.*;
import cs131w.parser.MyParser;
import cs131w.util.*;


public class Scope
{
	//----------------------------------------------------------------
	//	Instance variables.
	//----------------------------------------------------------------
		private Vector		m_lstLocals;
		private int			m_localSize;

	
	//----------------------------------------------------------------
	//
	//----------------------------------------------------------------
	public 
	Scope ()
	{
		m_lstLocals = new Vector ();
		m_localSize = 0;
	}


	//----------------------------------------------------------------
	//
	//----------------------------------------------------------------
	public STO
	access (String strName)
	{
		return	accessLocal (strName);
	}


	public Vector
	getMembers(){
		return m_lstLocals;
	}

	public STO
	getSingle(){
		return ((STO) m_lstLocals.firstElement());
	}

	protected void
	setMembers(Vector v){
		m_lstLocals = v;
	}


	public int
	getCount(){
		return  m_lstLocals.size();
	}


	public int
	getSize(){
		return m_localSize;
	}
	
	private void
	recalcSizeFromLocals(){
		m_localSize = 0;
		STO sto;
		
		for (int i = 0; i < m_lstLocals.size (); i++)
		{
			sto = (STO) m_lstLocals.elementAt (i);
			m_localSize += sto.getSize();
		}
	}

	//----------------------------------------------------------------
	//
	//----------------------------------------------------------------
	public STO
	accessLocal (String strName)
	{
		STO		sto = null;
		for (int i = 0; i < m_lstLocals.size (); i++)
		{
			sto = (STO) m_lstLocals.elementAt (i);
			if (sto.getName ().equals (strName))
				return (sto);
		}

		return (null);
	}

	//----------------------------------------------------------------
	//
	//----------------------------------------------------------------
	public void
	InsertLocal (STO sto)
	{
		m_lstLocals.addElement (sto);
		
		//if defining struct, but not func inside struct
		if (MyParser.getSymTab().isDefiningStruct() && !MyParser.getSymTab().isDefiningFunc()){
			MyParser.debug(this + ": adding var '" + sto +"' to struct, setting ofset as: "+m_localSize);
			//inc the offset by already set stuff
			sto.markStructMember(m_localSize);
			//and set that its already created, so it wont be again
			sto.setAsmblyCreated();
		}
		
		m_localSize += sto.getSize();
	}

	//----------------------------------------------------------------
	//
	//----------------------------------------------------------------

	@Override
	public Object 
	clone() { 	
		Scope ret = new Scope( );
		ret.setMembers((Vector)m_lstLocals.clone());
		ret.recalcSizeFromLocals();
		return ret;
	}


	public String
	toString(){
		if (!MyParser.DEV)
			return "";
		
		final StringBuilder sb = new StringBuilder();
		sb.append("Scope(#"+(m_lstLocals.size())+","+m_localSize+"b){");
		STO sto;
		
		for (int i = 0; i < m_lstLocals.size (); i++)
		{
			sto = (STO) m_lstLocals.elementAt (i);
			//commented out for now, because in case of pointer it recurses indefinite, and crashes
			sb.append(sto+",");
		}
		
		sb.append("}");
		return sb.toString();
		
	}

	public boolean 
	equals(Object obj){
		if (!(obj instanceof Scope)) return false;
		
		Scope o = (Scope) obj;
		
		if (o.getCount() != this.getCount())
			return false;
			
		Vector others = o.getMembers();
		
		STO		sto = null;
		STO		sto2 = null;
		for (int i = 0; i < m_lstLocals.size (); i++)
		{
			sto = (STO) m_lstLocals.elementAt (i);
			sto2 = (STO) others.elementAt (i);
			
			if ( !(sto.getName().equals(sto2.getName())) || !(sto.getParentType().isEquivalent(sto2.getParentType())) )
				return false;
		}
		
		return true;
		
	}

	//----------------------------------------------------------------
	//	Special method to be called when toString'ing a scope inside a struct
	//  prevents it from looping indefinitely if there is a pointer to itself within the struct
	//----------------------------------------------------------------
	public String
	toSpecialString(StructType current){
		if (!MyParser.DEV)
			return "";
		
		final StringBuilder sb = new StringBuilder();
		sb.append("Scope(#"+(m_lstLocals.size())+","+m_localSize+"b){");
		STO sto;
		
		for (int i = 0; i < m_lstLocals.size (); i++)
		{
			sto = (STO) m_lstLocals.elementAt (i);
			//do not append if its pointer to current struct
			if (sto.isVar() && sto.getParentType().isPtr() && ((PtrType)sto.getParentType()).getPointedToType() == current)
				sb.append("this" + Utils.strRepeat("*",((PtrType)sto.getParentType()).getDerefs()) + ",");
			
			//do the same if its an array of pointers
			else if (sto.isVar() && sto.getParentType().isArray() && sto.getParentType().getUnderlyingType().isPtr() && ((PtrType)sto.getParentType().getUnderlyingType()).getPointedToType() == current)
				sb.append("this" + Utils.strRepeat("*",((PtrType)sto.getParentType().getUnderlyingType()).getDerefs()) + "["+ ((ArrayType)sto.getParentType()).getDim() + "]" + ",");
			//or pointer to an array
			else if (sto.isVar() && sto.getParentType().isPtr() && sto.getParentType().getPointedToType().isArray() && ((PtrType)sto.getParentType()).getPointedToType().getUnderlyingType() == current)
				sb.append("this" + "["+ ((ArrayType)sto.getParentType().getPointedToType()).getDim() + "]" + Utils.strRepeat("*",((PtrType)sto.getParentType()).getDerefs()) + ",");
			
			else
				sb.append(sto+",");
		}
		
		sb.append("}");
		return sb.toString();
		
	}


	

}
