package cs131w.symtab;

//---------------------------------------------------------------------
//
//---------------------------------------------------------------------
import java.util.*;
//cs131 imports
import  cs131w.symtab.sto.*;
import  cs131w.symtab.types.*;
import  cs131w.parser.MyParser;


public class SymbolTable
{
	//----------------------------------------------------------------
	//	Instance variables.
	//----------------------------------------------------------------
		private Stack  		m_stkScopes;
		private int			m_nLevel;
		private Scope		m_scopeGlobal;
		private static STO	m_func = null;
		private static STO	m_struct = null;


	//----------------------------------------------------------------
	//
	//----------------------------------------------------------------
	public 
	SymbolTable ()
	{
		m_nLevel = 0;
		m_stkScopes = new Stack ();
		m_scopeGlobal = null;
	}


	//----------------------------------------------------------------
	//
	//----------------------------------------------------------------
	public void
	insert (STO sto)
	{
		Scope scope = (Scope) m_stkScopes.peek ();;
		scope.InsertLocal (sto);
	}


	//----------------------------------------------------------------
	//
	//----------------------------------------------------------------
	public STO
	accessGlobal (String strName)
	{
		return (m_scopeGlobal.access (strName));
	}

	//----------------------------------------------------------------
	//
	//----------------------------------------------------------------
	public STO
	accessLocal (String strName)
	{

		Scope		scope = (Scope) m_stkScopes.peek ();
		return (scope.accessLocal (strName));
	}


	public STO
	access (String strName)
	{
		return this.access(strName, 0);
	}
	
	public STO
	access (String strName, int start)
	{
		Stack		stk = new Stack ();
		Scope		scope;
		int n_start = (m_stkScopes.size()-1) - start;
		//int n_start = (m_stkScopes.size());
		
		STO			stoReturn = null;	

		//so search from the end, inside out
		for (int i = n_start; i >= 0 ; i-- ){
			scope = (Scope) m_stkScopes.get(i);
			if ((stoReturn = scope.access (strName)) != null)
				return	stoReturn;
			
		}

		return (null);
		
	}


	//----------------------------------------------------------------
	//
	//----------------------------------------------------------------
	public Scope
	openScope ()
	{
		Scope		scope = new Scope();

		//	The first scope created will be the global scope.
		if (m_scopeGlobal == null)
			m_scopeGlobal = scope;

		m_stkScopes.push (scope);
		m_nLevel++;
		return scope;
	}

	public Scope
	getInnermostScope ()
	{
		return (Scope) m_stkScopes.peek ();
	}


	//----------------------------------------------------------------
	//
	//----------------------------------------------------------------
	public void
	closeScope ()
	{
		m_stkScopes.pop ();
		m_nLevel--;
	}


	//----------------------------------------------------------------
	//
	//----------------------------------------------------------------
	public int
	getLevel ()
	{
		return m_nLevel;
	}


	public String
	toString(){
		if (!MyParser.DEV)
			return "";
		
		final StringBuilder sb = new StringBuilder();
		sb.append("ST("+m_nLevel+") ");
		if (isDefiningFunc())
			sb.append("of " + m_func);
		else if (isDefiningStruct())
			sb.append("of " + m_struct);
		
		//sb.append(m_stkScopes);
		
		return sb.toString();
	}

	//----------------------------------------------------------------
	//	This is the function currently being parsed.
	//----------------------------------------------------------------
	public STO					getFunc () { return m_func; }
	public void					setFunc (STO sto) { m_func = sto; }
	public static boolean		isDefiningFunc () { return (m_func != null); }

	//----------------------------------------------------------------
	//	This is the struct currently being parsed.
	//----------------------------------------------------------------
	public STO					getStruct () { return m_struct; }
	public void					setStruct (STO s) { m_struct = s; }
	public static boolean		isDefiningStruct () { return (m_struct != null); }








}
