package cs131w.symtab.ops;

//---------------------------------------------------------------------
// This is the top of the Type hierarchy. You most likely will need to
// create sub-classes (since this one is abstract) that handle specific
// types, such as IntType, FloatType, ArrayType, etc.
//---------------------------------------------------------------------
import  cs131w.lexer.Token;
import	cs131w.parser.MyParser;
import cs131w.symtab.ops.unary.UnaryIncDecOp;
import cs131w.symtab.ops.binary.*;
import  cs131w.symtab.sto.*;
import  cs131w.assembly.*;


public abstract class Operator implements Asmbl 
{

	//the internal name of the opp
	protected String  		m_typeName;
	//the corresponding token
	protected Token  		m_Token;
	//if there is a template to save some code for sub functions of asmblyDoOp
	protected String		m_Template;

	protected String		m_FloatTemplate;
	//----------------------------------------------------------------
	//
	//----------------------------------------------------------------
	public 
	Operator ()
	{
		setName("op");
		setToken(null);
	}


	public 
	Operator (String strName, Token tok)
	{
		setName(strName);
		setToken(tok);
	}


	//----------------------------------------------------------------
	//
	//----------------------------------------------------------------
	public String
	getName ()
	{
		return m_typeName;
	}

	private void
	setName (String str)
	{
		m_typeName = str;
	}

	public Token
	getToken ()
	{
		return m_Token;
	}

	private void
	setToken (Token t)
	{
		m_Token = t;
	}
	
	//protected boolean isModifyingOp(){ return false; }
	
	//----------------------------------------------------------------
	//	functions that check operand/s of this op
	//----------------------------------------------------------------
	public STO 
	checkOperands(STO a, STO b){
		return a;
	}
		
	public STO 
		checkOperand(STO a){
			return a;
	}
		
	//----------------------------------------------------------------
	//	functions that apply operand/s of this op
	//----------------------------------------------------------------
	public STO operate(STO a, STO b){
		a.setRVal();
		return a;
	}

	public STO operate(ConstSTO a, ConstSTO b){
		return operate((STO)a, (STO)b);
	}

	public STO operate(STO a){
		MyParser.debug(a + " set to R-value in Operator, make sure you want that");
		a.setRVal();
		return a;
	}

	public STO operate(ConstSTO a){
		return operate((STO)a);
	}
	
	//----------------------------------------------------------------
	//
	//----------------------------------------------------------------
	public String
	toString()
	{
		return m_typeName;
	}	
	

	// -----------------------------------------------------------------
	//----------------------------------------------------------------
	//	Assembly functions
	//	these emit the necc. assembly code for doing prescribed action
	//----------------------------------------------------------------

	//this does the op in asmbly
	//a and b are both actual operands, res is the already processed result from the 'normal' operate above
	//ie. has right type and flags
	//loads both ops, calls the DoOp method that should be overwritten by each op and stores the result back
	public STO
	AsmblyOperate(STO res, STO a, STO b){ 
		//do NOT do this if both ops are Const, in this case it will be done at compile time,
		//by the overriden operate(const, const)
		if (a.isConst() && b.isConst()) return res;
		
		MyParser.debug(this + ".AsmblyOp(a,b) called!");
		MyParser.asmblyWriteVarSmt(COMMENT, "--- Begin "+this+" Op ---");
		
		a.AsmblyClearReg();
		b.AsmblyClearReg();
		
		a.AsmblyLoad();
		b.AsmblyLoad();
		
		//then do the opration
		STO r = AsmblyDoOp(res, a, b);

		r.AsmblyStore();
		
		
		//and clear both used ops from regs
		a.AsmblyClearReg();
		b.AsmblyClearReg();
		
		MyParser.asmblyWriteVarSmt(COMMENT, "--- End "+this+" Op ---");
		//and return it
		return r;
	}
	
	public STO
	AsmblyOperate(STO res, STO a){ 
		//do not do this if const
		if (a.isConst()) return res;
		
		MyParser.asmblyWriteVarSmt(COMMENT, "--- Begin "+this+" Op ---");
		
		MyParser.debug(this + ".AsmblyOp(a,b) called!");
		//this needs to emit code to load both operands
		a.AsmblyLoad();
		//res.AsmblyLoad(); // unary inc/dec uses this, don't remove, not redundant
		//then do the operation
		AsmblyDoOp(res, a);
		//and store the result back 
		a.AsmblyStore();
		res.AsmblyStore();

		MyParser.asmblyWriteVarSmt(COMMENT, "--- End "+this+" Op ---");
		
		return res;
	}
	
	//this needs to be overwritten by each binary op
	//a and b are both operands, res is the already created temporary ExprSTO
	public STO
	AsmblyDoOp(STO res, STO a, STO b){
		return res;
	} 
	

	//this needs to be overwritten by each unary op
	public STO
	AsmblyDoOp(STO res, STO a){
		MyParser.debug(this + ".AsmblyDoOp(a) called. OVERRIDE ME!");
		return res;
	} 

	

}