package cs131w.symtab.ops.binary;

//---------------------------------------------------------------------
// basic math ops, like +-*/
//---------------------------------------------------------------------
import  cs131w.symtab.ops.Operator;
import  cs131w.lexer.*;
import  cs131w.parser.*;
import  cs131w.symtab.sto.*;
import  cs131w.symtab.ops.unary.*;
import  cs131w.symtab.types.*;
import  cs131w.error.*;
import  cs131w.assembly.*;


public class AssignOp extends BinaryOp{
	
	
	//----------------------------------------------------------------
	// ctor
	//----------------------------------------------------------------
	public 
	AssignOp ()
	{
		super("=", new Token (sym.T_ASSIGN, "="));
		
	}
	
		
	//function that checks operand of this op
	public STO 
		checkOperands(STO stoTarget, STO stoSrc){

			MyParser.debug(this + ": checking " + stoTarget + " = " + stoSrc);

			//check if target is ModLVal
			if (!stoTarget.isModLValue())
			{
				MyParser.debug(this + ": checking target, not mod LVal");
				return new ErrorSTO (stoTarget, ErrorMsg.error3a_Assign);
			}
			
			//check if we are assigning struct = struct
			if (stoSrc.isStruct() && stoTarget.isStruct() && stoTarget.getParentType().getRootType().getClazz() != stoSrc.getParentType().getRootType().getClazz()){
				MyParser.debug(this + ": checking struct equiv, tar.t:" + stoTarget.getType() + ", s.t: " + stoSrc.getType());
				return new ErrorSTO (stoSrc, Formatter.toString(ErrorMsg.error3b_Assign, ((StructType)stoSrc.getType()).getTag(), ((StructType)stoTarget.getType()).getTag()));
			}
			
			//check target is funcPtr = func
			//see if its an overloaded func, we need to check if the right overloaded signature is available
			if ( (stoTarget.getParentType().isFuncPtr() && stoSrc.isUeberFunc())){ 
				MyParser.debug(this + ": checking funcPtr = overloaded func...");
				//check if there is an overloaded func with the right signature
				
				//get the sig from the target ptr
				String mangled = stoTarget.getParentType().getRootType().getPointedToType().getMangledName();
				STO func;
				//now check if there is an overloaded func with that sig in the src ueberSTO
				if ( (func = ((UeberFuncSTO)stoSrc).access(mangled)) == null){
					return new ErrorSTO (stoSrc, Formatter.toString(ErrorMsg.error3b_Assign, stoSrc.getType(), stoTarget.getType()));
				}
				
				//yes that means there is, now we need to check it rigurously, so assign the func to Src and fall thru to next check
				stoSrc = func;
				
			}
			
			//check if target is a function pointer and src a func thats assignable
			if ( (stoTarget.getParentType().isFuncPtr() && stoSrc.getParentType().isFunction())){ 
					MyParser.debug(this + ": checking funcPtr = func...");
					
					if (stoTarget.getParentType().getRootType().getPointedToType().isEquivalent(stoSrc.getParentType().getRootType()))
						//this works, so we want to just return
						return stoTarget;
					else
						return new ErrorSTO (stoSrc, Formatter.toString(ErrorMsg.error3b_Assign, ((FuncSTO)stoSrc).getProtoType(), stoTarget.getType()));
			}
			
			//check if both are funcPtrs
			if ( stoTarget.getParentType().isFuncPtr() && stoSrc.getParentType().isFuncPtr() ){ 
				MyParser.debug(this + ": checking funcPtr = funcPtr...");
				
				if (stoTarget.getParentType().getRootType().getPointedToType().isEquivalent(stoSrc.getParentType().getRootType().getPointedToType()))
					//this works, so we want to just return
					return stoTarget;
				else
					return new ErrorSTO (stoSrc, Formatter.toString(ErrorMsg.error3b_Assign, stoSrc.getType(), stoTarget.getType()));

			}

			//none of the above, just check basic assignabilty
			MyParser.debug(this + ": checking basic assignability of " + stoTarget.getParentType() + " = " + stoSrc.getParentType() + ": " + stoSrc.getParentType().isAssignable(stoTarget.getParentType()));
			if (!stoSrc.getParentType().isAssignable(stoTarget.getParentType())){
				
				return new ErrorSTO (stoSrc, Formatter.toString(ErrorMsg.error3b_Assign, stoSrc.getType(), stoTarget.getType()));

			}
			
			MyParser.debug(this + ": returning " + stoTarget);
			return stoTarget;
			
		}
	
	@Override
	public STO operate(STO stoTarget, STO stoSrc){
		MyParser.debug(this + ".operate: '" +stoTarget+ "' = " + (stoSrc));
		
		//do nothing
		return stoTarget;
	}


	//this needs to be overwritten by each binary op
	public STO
	AsmblyDoOp(STO res, STO trg, STO src){

		//check if we are assigning a func or ueberfunc to a funcPtr
		if ( trg.getParentType().isFuncPtr() && 
			(src.isUeberFunc() || src.getParentType().isFunction() )
			){ 
		
			return doFuncAss(res, trg, src);

		}
		//if we are assining a funcPtr to a funcPtr we do not need to do anything, since they are both pointers
		
		//first check if we need to promote the src to float
		if (trg.getType().isFloat() && src.getType().isInt()){
	
			//promote src to float by using the cast op
			src = MyParser.DoBinaryExpr(new EmptySTO(new FloatType()), new CastOp(), src);
			//and load back up
			src.AsmblyLoad();
			//WARNING: now we are also responsible for freeing it from reg after using since we have changed the instance of src
			//and the calling AsmblyOp in Operator can't free this new instance, since it has no way of knowing
			//it will still try to clrReg the old instance, so we do it ourselves at the end
		}
		
		//check if we assining the name of an array to a pointer
		if (src.getType().isArray() && trg.getType().isBasicPtr()){
			//in this case we need to just load the addr of the array, rather than the first value, which is what happens otherwise
			src.AsmblyLoadAddr(src.getReg());
		}
		
		//no we do the actual op in asmbly
		String op = MOVE;
		
		//this needs to be written thru MyParser, coz it could be in global INIT
		if (src.getType().isFloat() || trg.getType().isFloat())
			op = FMOVE;
			
		MyParser.asmblyWriteVarSmtComm (this+": " + trg.getReg() + " = " + src.getReg(), op, src.getReg(), trg.getReg());
		
		//now the src is in trg
		//we can clear the orig. src reg
		src.AsmblyClearReg();
		
		//and return the res expr STO which IS trg anyway
		return res;
	} 

	//does the assignment from a function in Src to a funcPtr in Trg
	private STO
	doFuncAss(STO res, STO stoTarget, STO stoSrc){
		
		STO func = stoSrc;
		
		if ( stoSrc.isUeberFunc()) { 
			
			//get the sig from the target ptr
			String mangled = stoTarget.getParentType().getRootType().getPointedToType().getMangledName();
			
			//now get the right overloaded func from ueberFunc
			func = ((UeberFuncSTO)stoSrc).access(mangled);
	
		}
		
		//now we just set the address of this func into a reg
		MyParser.asmblyWriteVarSmtComm (this+": assgning func to FuncPtr", SET, func.getLabel(), res.getReg());
		
		//and check if this is part of a struct, we store a ptr to that struct in a tempMemLoc
		if (stoSrc.isStructMember()){
			
			//we need to load a ptr to that struct first, do it into L7
			stoSrc.m_memberOfStruct.AsmblyLoadAddr(L7);
			
			//and signal that the pointer is in L7, which cases it to make a new temp location and store that ptr there
			res.setStructThisPtr(new StorageLocation(L7));
			
		}
		
		
		return res;
	}

}