package cs131w.symtab.ops.binary;

//---------------------------------------------------------------------
// basic math ops, like +-*/
//---------------------------------------------------------------------
import cs131w.symtab.ops.Operator;
import  cs131w.lexer.*;
import  cs131w.parser.*;
import  cs131w.symtab.sto.*;
import  cs131w.symtab.types.*;
import  cs131w.error.*;


public abstract class BasicOp extends BinaryOp{
	
	//----------------------------------------------------------------
	// ctor
	//----------------------------------------------------------------
	public 
	BasicOp ()
	{
		super("", null);
	}
	

	public 
	BasicOp (String strName, Token tok)
	{
		super(strName, tok);
	}
	
	//this function needs to check both operands if they are equal or coercible and return appr type
	public STO 
	checkOperands(STO a, STO b) { 
		MyParser.debug(this + ": checking " + a + ", and " + b);
		
		if ( !a.getType().isNumeric()){
			
			return (new ErrorSTO( Formatter.toString(ErrorMsg.error1n_Expr, a.getType(), this)));
			
		}else if ( !b.getType().isNumeric()) { 
		
			return (new ErrorSTO( Formatter.toString(ErrorMsg.error1n_Expr, b.getType(), this)));
			
		}
		
		return a;
	}
	
	//this applies the actual operator to the types and returns appropriate type for the ops onvolved
	//if they are of type ConstSTO, the operator can already applied to the underlying value, this will be overriden by the op
	//runtime system ensures that appr. method will be called
	public STO 
	operate(STO a, STO b){
		STO res;
		MyParser.debug(this + ".operate: '" +a+ "', " + b);
		
		if (a.getType().isInt() && b.getType().isInt()) {
			//return new ExSTO of type Integer
			res = (new ExprSTO("basicop", new IntType()));
		} else {
			res = (new ExprSTO("basicop", new FloatType()));
		}
		
		res.setRVal();
		return res;
	}
	
	//----------------------------------------------------------------
	// Asmbly
	//----------------------------------------------------------------
	
	//this needs to be overwritten by each binary op
	//here we check what type of op we need to execute (write code for)
	public STO
	AsmblyDoOp(STO res, STO a, STO b){
		STO ret = res;
		boolean a_changed = false;
		boolean b_changed = false;
		
		//first check if we need to promote one of them to float
		if (res.getType().isFloat() && a.getType().isInt()){
	
			//promote a to float by using the cast op
			a = MyParser.DoBinaryExpr(new EmptySTO(new FloatType()), new CastOp(), a);
			//and load back up
			a.AsmblyLoad();
			//now we are also responsible for freeing it from reg after using since we have chaged the instance of src
			//and the calling AsmblyOp in Operator can't free this new instance, since it has no way of knowing
			//it will still try to clrReg the old instance
			a_changed = true;
			
		}else if (res.getType().isFloat() && b.getType().isInt()){
			
			//promote b to float by using the cast op
			b = MyParser.DoBinaryExpr(new EmptySTO(new FloatType()), new CastOp(), b);
			//and load back up
			b.AsmblyLoad();
			//now we are also responsible for freeing it from reg after using since we have chaged the instance of src
			//and the calling AsmblyOp in Operator can't free this new instance, since it has no way of knowing
			//it will still try to clrReg the old instance
			b_changed = true;
		
		}
		
		
		//here we need to check first if we are doing an fp op
		if (res.getType().isFloat()){
			ret = AsmblyDoFlOp(res,a,b);
		}else{
			//integer: now check if its add/sub or mul/div
			if (m_typeName.equals("+") || m_typeName.equals("-"))
				ret = AsmblyDoAddOp(res,a,b);
			else
				ret = AsmblyDoMulOp(res,a,b);
		}
		
		//now the src is in trg
		//we can clear the orig. src reg
		if (a_changed)
			a.AsmblyClearReg();
		
		if (b_changed)
			b.AsmblyClearReg();
			
		return ret;
	}

	//doing the op for add/sub ops integer
	public STO
	AsmblyDoAddOp(STO res, STO a, STO b){
		//no we do the actual op in asmbly
		//this needs to be written thru MyParser, coz it could be in global INIT
		MyParser.asmblyWriteVarSmtComm (this+": L0 = " + a.getReg() + " (+) " + b.getReg(), m_Template, a.getReg(), b.getReg(), L0);
		
		//and signal the res STO that it is stored in L0
		res.setRegister(L0);
		//and return the res expr STO
		return res;
	} 

	//mul/div integer
	public STO
	AsmblyDoMulOp(STO res, STO a, STO b){
		//no we do the actual op in asmbly
		//this needs to be written thru MyParser, coz it could be in global INIT
		MyParser.asmblyWriteVarSmtComm ("move op a into out reg", MOVE, a.getReg(), O0);
		MyParser.asmblyWriteVarSmtComm ("move op b into out reg", MOVE, b.getReg(), O1);
		MyParser.asmblyWriteVarSmtComm (this+": L0= a (*) b", CALL, m_Template);
		MyParser.asmblyWriteVarSmtComm ("delay slot", NOP);
		
		MyParser.asmblyWriteVarSmtComm ("move res into L0", MOVE, O0, L0);
		//and signal the res STO that it is stored in L0
		res.setRegister(L0);
		//and return the res expr STO
		return res;
	} 


	//float ops
	public STO
	AsmblyDoFlOp(STO res, STO a, STO b){
		//no we do the actual op in asmbly
		//this needs to be written thru MyParser, coz it could be in global INIT
		MyParser.asmblyWriteVarSmtComm (this+": F0 = " + a.getReg() + " (*) " + b.getReg(), m_FloatTemplate, a.getReg(), b.getReg(), F0);
		
		//and signal the res STO that it is stored in F0
		res.setRegister(F0);
		//and return the res expr STO
		return res;
	} 
	
	
}