package cs131w.symtab.ops.binary;

//---------------------------------------------------------------------
// basic math ops, like +-*/
//---------------------------------------------------------------------
import  cs131w.symtab.ops.Operator;
import  cs131w.lexer.*;
import  cs131w.parser.*;
import  cs131w.assembly.*;
import  cs131w.symtab.sto.*;
import  cs131w.symtab.ops.unary.*;
import  cs131w.symtab.types.*;
import  cs131w.error.*;
import  cs131w.stmts.*;
import cs131w.util.Utils;


public class BracketOp extends BinaryOp{
	
	
	//----------------------------------------------------------------
	// ctor
	//----------------------------------------------------------------
	public 
	BracketOp ()
	{
		super("[]", new Token (sym.T_LBRACKET, "["));
		
	}
		
	//function that checks operand of this op
	public STO 
		checkOperands(STO arr, STO idx){

			MyParser.debug(this + ": checking " + arr + ", " + idx);
			//this basically just needs to do a star and then a dot op in order
			
			//not array or pointer type
			if ( !arr.getType().isArray() && !arr.getType().isPtr()){
				return new ErrorSTO (arr, Formatter.toString(ErrorMsg.error11t_ArrExp, arr.getType()));
			}

			//index not int
			if ( !idx.getType().isInt()){
				return new ErrorSTO (idx, Formatter.toString(ErrorMsg.error11i_ArrExp, idx.getType()));
			}

			//if not pointer type and idx is Const, check bounds
			if (arr.getType().isArray() && idx.isConst()){

				ArrayType a = (ArrayType) (arr.getParentType().getRootType()) ;
				
				int i = ((ConstSTO)idx).getIntValue();

				//debug("arr.t: " + arr.getType());
				//debug("arr.pt: " + arr.getParentType());
				//debug("arr.pt.rootT: " + arr.getParentType().getRootType());

				int dim =  a.getDim();

				if (i < 0 || i > (dim-1)){
					return new ErrorSTO (arr, Formatter.toString(ErrorMsg.error11b_ArrExp, i, dim));
				}
			}
			
			return arr;
			
		}
	
	@Override
	public STO operate(STO arr, STO idx){
		MyParser.debug(this + ".operate: '" +arr+ "', " + (idx));
		
		ArrayType a = null;
		PtrType p = null;
		ExprSTO base;
		
		if (arr.getType().isArray()){
			a = (ArrayType) (arr.getParentType().getRootType()) ;
			base = new ExprSTO(arr+"["+idx+"]", a.getUnderlyingType());
		}else{
			p = (PtrType) (arr.getParentType().getRootType()) ;
			base = new ExprSTO(arr+"["+idx+"]", p.getPointedToType());
		}
		
		//if either array or struct
		if (base.getParentType().isArray() || base.getParentType().isStruct()){
			base.setIsLocal(arr.isLocal());
		}
		
		if (!base.getParentType().isArray())
			base.setModLVal();
		
		MyParser.debug(this + ".operate: returning '" +base+ "'");
		return base;
		
	}

	//----------------------------------------------------------------
	//	Assembly functions
	//	these emit the necc. assembly code for doing prescribed action
	//----------------------------------------------------------------


	//accessing specific el of array
	public STO
	AsmblyDoOp(STO res, STO arr, STO idx){
		MyParser.debug(this + ".AsmblyOp: '" +arr+ "'[" + idx + "] = " + res);
		
		//no we do the actual op in asmbly
		//multiply by size of underlying element, first get its size
		int uSize = arr.getType().getRootType().getUnderlyingType().getSize();
		
		//MyParser.debug(this + ".AsmblyOp: uSize("+(arr.getParentType().getUnderlyingType())+") is: " + uSize);
		
		boolean zero_offset = false;
		
	    //array bounds check
        MyParser.asmblyWriteVarSmt (COMMENT, "--- Begin array bounds checking ---");
        int dim = -1;
        if (arr.getType() instanceof ArrayType){
        	dim = ((ArrayType)arr.getType()).getDim();
        }else if(arr.getType().isAlias()){
        	if(((AliasType)arr.getType()).getRootType().isArray()){
        		dim = ((ArrayType)(((AliasType)arr.getType()).getRootType())).getDim();
        	}
        }
        if(dim >=0){        	
        	ConstSTO dimSTO = new ConstSTO(dim);
        	dimSTO.AsmblyCreate();
        	if(!dimSTO.isInReg()){
        		dimSTO.AsmblyLoad();
        	}
        	if(!idx.isInReg()){
        		idx.AsmblyLoad();
        	}

        	MyParser.asmblyWriteVarSmtComm("Compare index to dim of array", CMP, dimSTO.getReg(), idx.getReg() );
        	String failLabel = Utils.getSomeLabel("bracket_fail");
        	String okLabel = Utils.getSomeLabel("bracket_ok");
        	MyParser.asmblyWriteVarSmtComm("If dim<=idx, fail", BLE, failLabel);
        	MyParser.asmblyWriteVarSmt(NOP);
        	MyParser.asmblyWriteVarSmtComm("Compare index to 0", CMP, idx.getReg(), "0" );
        	
        	MyParser.asmblyWriteVarSmtComm("If idx>0, we're OK", BGE, okLabel);
        	MyParser.asmblyWriteVarSmt(NOP);
        	MyParser.asmblyWriteVarSmtComm("Begin fail code", LABEL, failLabel);
        	
			/*
			ConstSTO errorMsg = new ConstSTO("Index value of %d is outside legal range [0,%d).\\n");
        	errorMsg.AsmblyCOut();
			errorMsg.AsmblyCreate();
        	*/
			
			MyParser.asmblyWriteVarSmt (SET, "ArrayIdxE", O0);
			MyParser.asmblyWriteVarSmt (MOVE, idx.getReg(), O1);
        	MyParser.asmblyWriteVarSmt (MOVE, dimSTO.getReg(), O2);
        	MyParser.asmblyWriteVarSmt(CALL, "printf");
			
			MyParser.asmblyWriteVarSmt(CALL, "exit");
        	MyParser.asmblyWriteVarSmt(NOP);
        	MyParser.asmblyWriteVarSmt(RET);
        	MyParser.asmblyWriteVarSmt(RESTORE); 
        	MyParser.asmblyWriteVarSmtComm("If the array access was OK, go on from here", LABEL, okLabel);
        	MyParser.asmblyWriteVarSmt (COMMENT, "--- End array bounds checking ---");
        	dimSTO.AsmblyClearReg();
        	
        }
    	
        //see if it is const, and maybe 0
		
		if (idx.isConst() && ((ConstSTO)idx).getIntValue() == 0){
			zero_offset = true;
			//move to L0 
			//MyParser.asmblyWriteVarSmtComm ("move idx into L0", MOVE, idx.getReg(), L0);
		}else{
			//check if thats an exact multiple of 2
			boolean mult2 = ((uSize-1) & uSize) == 0;

			if (mult2){
				//now shift by that size ld_2
				long shift_by = Math.round(Math.log(uSize) / Math.log(2));
			
				AssemblyWriter.writeSmtAndComment (idx.getRegister() + " = (idx*baseSize("+uSize+") ) => offset", SLL, idx.getRegister(), new Long( (shift_by) ).toString(), L0);

			}else{
				//if not exact multiple, we just do vanilla multiply
				MyParser.asmblyWriteVarSmtComm ("move op a into out reg", MOVE, idx.getReg(), O0);
				MyParser.asmblyWriteVarSmtComm ("move op b into out reg", SET, new Integer(uSize).toString(), O1);
				MyParser.asmblyWriteVarSmtComm (this+": L0= a (*) b", CALL, ".mul");
				MyParser.asmblyWriteVarSmtComm ("delay slot", NOP);
				//move from out reg into L0
				MyParser.asmblyWriteVarSmtComm ("move res into L0", MOVE, O0, L0);
	
			}
		}
		
				
		ArrayType a = (ArrayType) (arr.getParentType().getRootType()) ;
		
		//the base addr of arr is already loaded, oh well! just do it again
		//AssemblyWriter.writeSmtAndComment ("set addr of '" + arr + "' into L7", SET, arr.getMemBase(), L7);
		arr.AsmblyLoadAddr(L7);
		
		if (!zero_offset){
			//now add the offset to base which is in L7
			//positive if global
			//if (!arr.isLocal())
			//always pos.
			MyParser.asmblyWriteVarSmtComm ("L7 = Base + offset", ADD, L7, L0, L7);
			//else	//or neg if local
			//	AssemblyWriter.writeSmtAndComment ("L7 = (local) Base - offset", SUB, L7, L0, L7);

		}
		
		//now just let the ExprSTO know that it has a runtime addr, which is in L7
		res.setRuntimeAddr(new StorageLocation(L7));
		//this is going to make it store that runtime addr somewhere in temp mem
		//and then use that addr to to load and store the res' value from there, instead of from some bogus location
		
		//and load to there
		String tar = L0;
		
		if (a.getUnderlyingType().isFloat()){
			tar = F0;
		}	
		
		MyParser.asmblyWriteVarSmtComm ("now load from [L7] into " + tar, LOAD, L7, tar);
		//signal that the result, ie. element of array is now in L0 or F0
		res.setReg(tar);
		idx.AsmblyClearReg();
	
		return res;
	} 
	
	

}
