package cs131w.symtab.ops.binary;

//---------------------------------------------------------------------
// basic math ops, like +-*/
//---------------------------------------------------------------------
import 	cs131w.symtab.ops.Operator;
import  cs131w.lexer.*;
import  cs131w.parser.*;
import  cs131w.symtab.sto.*;
import  cs131w.symtab.types.*;
import  cs131w.error.*;
import  cs131w.assembly.*;


public class CastOp extends BinaryOp{
	
	//----------------------------------------------------------------
	// ctor
	//----------------------------------------------------------------
	public 
	CastOp ()
	{
		super("()", new Token (sym.T_LPAREN, "("));
	}
		
	//function that checks operand of this op
	public STO 
		checkOperands(STO type, STO sto){
			MyParser.debug(this + ": checking ( "+type+" )" + sto);
			//get the root type out of the empty sto type
			Type t = type.getType();
			
			if ( !sto.getParentType().isCastable(t.getRootType())){
				return new ErrorSTO (sto, Formatter.toString(ErrorMsg.error20_Cast, sto.getType(), t));
			}		
			
			return sto;
			
		}
	
	@Override
	public STO operate(STO type, STO sto){
	
		MyParser.debug(this + ".operate: '" +type+ "', " + (sto));
		Type t = type.getType();
		STO ret;
		//if all is well, all we have to do now is make a copy of the original STO and change its type
		
		ret = (STO) sto.clone();
		ret.setType(t);
		ret.setRVal();
		//and set that this needs to be created anew
		ret.setAsmblyCreated(false);
		
		return ret;
	}
	
	
	
	public STO
	AsmblyDoOp(STO res, STO type, STO a){
		
		MyParser.asmblyWriteVarSmt (COMMENT, "--- Begin Cast Op " +a.getType()+ " -> " +type.getType()+ " ---");
		
		//convert int/bool -> float
		if ((a.getType().isInt() || a.getType().isBool()) && type.getType().isFloat()){
			
			//first get a temp location
			int off = MyParser.increaseFrameOffset();
			String flreg = Regs.getNextFloatReg();
			
			MyParser.asmblyWriteVarSmtComm ("store "+a+" into temp mem location", STORE, a.getRegister(), "%fp"+off);
			MyParser.asmblyWriteVarSmtComm ("load back into FP reg", LOAD, "%fp"+off, flreg);
			MyParser.asmblyWriteVarSmtComm ("and convert", FITOS, flreg, flreg);
			
			//Regs.FreeReg(a.getReg());
			
			//and signal res that its in flreg
			res.setReg(flreg);
			
		}
		
		//convert float -> int
		if(a.getType().isFloat() && type.getType().isInt()){
			
			//first get a new reg
			String reg = Regs.getNextReg();
			int off = MyParser.increaseFrameOffset();
			
			MyParser.asmblyWriteVarSmtComm ("convert single precision float -> int", FSTOI, a.getReg(), a.getReg());
			//AssemblyWriter.writeSmtAndComment ("and move", FMOVE, a.getReg(), reg);
			MyParser.asmblyWriteVarSmtComm ("store "+a+" into temp mem location", STORE, a.getReg(), "%fp"+off);
			MyParser.asmblyWriteVarSmtComm ("load back into Int reg", LOAD, "%fp"+off, reg);
			
			//Regs.FreeReg(a.getReg());
			
			res.setReg(reg);
		}
		
		//convert int -> bool
		if (a.getType().isInt() && type.getType().isBool()){
			String trueLabel = AssemblyWriter.getNextCmpTempLabel();
			String endLabel = AssemblyWriter.getNextCmpTempLabel();
			res.AsmblyLoad();
			MyParser.asmblyWriteVarSmtComm("CastOp: Compare " + a + "to 0", CMP , a.getRegister(), "0");
			MyParser.asmblyWriteVarSmt(BNE, trueLabel);
			MyParser.asmblyWriteVarSmt(NOP);
			MyParser.asmblyWriteVarSmt(SET, "0", L0);
			MyParser.asmblyWriteVarSmt(BA, endLabel);
			MyParser.asmblyWriteVarSmt(NOP);
			MyParser.asmblyWriteVarSmt(LABEL, trueLabel);
			MyParser.asmblyWriteVarSmt(SET, "1", L0);
			MyParser.asmblyWriteVarSmt(LABEL, endLabel);
			
			res.setReg(L0);
			
		}

		//convert bool-> int
		if(a.getType().isBool() && type.getType().isInt()){
			res.setReg(a.getReg());
		}
		//convert float -> bool;
		if(a.getType().isFloat() && type.getType().isBool()){
			String trueLabel = AssemblyWriter.getNextCmpTempLabel();
			String endLabel = AssemblyWriter.getNextCmpTempLabel();
			res.AsmblyLoad();
			String flreg = Regs.getNextFloatReg();
			ConstSTO zero = new ConstSTO((float)0.0);
			zero.setType(new FloatType());
			zero.AsmblyCreateTemp();
			zero.AsmblyLoad();
			MyParser.asmblyWriteVarSmtComm("CastOp: Compare " + a + "to 0.0", FCMP , a.getRegister(), zero.getReg());
			MyParser.asmblyWriteVarSmt(FBNE, trueLabel);
			MyParser.asmblyWriteVarSmt(NOP);
			MyParser.asmblyWriteVarSmt(SET, "0", L0);
			MyParser.asmblyWriteVarSmt(BA, endLabel);
			MyParser.asmblyWriteVarSmt(NOP);
			MyParser.asmblyWriteVarSmt(LABEL, trueLabel);
			MyParser.asmblyWriteVarSmt(SET, "1", L0);
			MyParser.asmblyWriteVarSmt(LABEL, endLabel);
			zero.AsmblyClearReg();
			res.setReg(L0);
		}
		
		MyParser.asmblyWriteVarSmt (COMMENT, "--- End Cast Op " +a.getType()+ " -> " +type.getType()+ " ---");
	
		return res;
	
	}

}