package cs131w.symtab.ops.binary;

//---------------------------------------------------------------------
// basic math ops, like +-*/
//---------------------------------------------------------------------
import  cs131w.symtab.ops.Operator;
import  cs131w.lexer.*;
import  cs131w.parser.*;
import  cs131w.assembly.*;
import  cs131w.symtab.sto.*;
import  cs131w.symtab.types.*;
import  cs131w.error.*;


public class DotOp extends BinaryOp{
	
	//----------------------------------------------------------------
	// ctor
	//----------------------------------------------------------------
	public 
	DotOp ()
	{
		super(".", new Token (sym.T_DOT, "."));
	}
		
	//function that checks operand of this op
	public STO 
		checkOperands(STO a, STO b){

			MyParser.debug(this + ": checking " + a + ", and name " + b);
			
			String strID = b.getName();
			//String func_mang = strID + "#";
			// Good place to do the struct checks
			if (!(a.getType().isStruct()) )
			{
				return new ErrorSTO (Formatter.toString(ErrorMsg.error14t_StructExp, a.getType()));
			}

			//now see if the struct has that member
			StructType struct = (StructType) (a.getParentType().getRootType());

			if ((struct.access (strID)) == null) 			//&& !struct.prefixExists (func_mang))
			{
			 	//see if we are this'ing
				if (MyParser.getSymTab().isDefiningStruct() && (MyParser.getSymTab().getStruct() == a))
					return new ErrorSTO (Formatter.toString(ErrorMsg.error14b_StructExpThis, strID));	
				else
					 return new ErrorSTO (Formatter.toString(ErrorMsg.error14f_StructExp, strID, struct));	
			}

			return a;
			
		}
	
	@Override
	public STO operate(STO a, STO b){
		//yes it is, so return a new ExprSTO with this pointer deref'fed once
		MyParser.debug("DotOp.opearate: sto '" +a+ "'.name= " + (b));
		
		String strID = b.getName();
		//String func_mang = strID + "#";
		
		//now find the appr member of the struct and return it
		StructType struct = (StructType) (a.getParentType().getRootType());
		//now check if this is a overloaded function inside the struct, in this case we just return the struct rather than the member, since we cant decide that yet
		//if (struct.prefixExists (func_mang))
		//	return a;
			
		STO s = struct.access (strID);
		
		//here we need to create a new ExprSTO for this var, if its not a func
		if (!s.isFunc()){
			ExprSTO mem = new ExprSTO(a.getName()+"."+b.getName(), s.getParentType());
			mem.setModLVal();
			//since this is going to get a runtime addr. assigned later, we just copy the original mem_offset over from the structs member
			//which represents the right offset from the base addr. of the struct
			mem.setMemOffset(s.getMemOffset());
			
			//if parent type is global make sure that we copy that flag
			mem.setIsLocal(a.isLocal());
			
			return mem;
		}else if(s.isFunc()){
			
			//this is func, so return that element instead
			//and signal that this is a struct mem
			s.m_isStructMember = true;
			s.m_memberOfStruct = a;
			
			MyParser.debug("DotOp.opearate: returning struct field as func: " + s);
			return s;
		
		}
		
		return s;	
	}



	//accessing specific element of struct
	public STO
	AsmblyDoOp(STO res, STO s, STO id){
		MyParser.debug(this +"DotOp.AsmblyOperate: res='" +res+ "', s= " + (s) + ", id=" + id);
		
		//StructType struct = (StructType) (s.getParentType().getRootType());
		//STO mem = struct.access (id.getName());
		
		//res is already the corresponding STO from the struct's scope
		//just get the appr. offset from it

		//the base addr of struct is already loaded, oh well! just do it again
		//AssemblyWriter.writeSmtAndComment ("set addr of '" + s + "' into L7", SET, s.getMemBase(), L7);
		s.AsmblyLoadAddr(L7);
		
		//now add the offset
		//positive if global
		//if (!s.isLocal())
		//always positive
		AssemblyWriter.writeSmtAndComment ("L7 = Base + offset of "+id.getName()+" ("+res.getMemOffset()+")", ADD, L7, new Integer(res.getMemOffset()).toString(), L7);
		//else	//or neg if local
		//	AssemblyWriter.writeSmtAndComment ("L7 = (local) Base - offset of "+id.getName()+" ("+res.getMemOffset()+")", SUB, L7, new Integer(res.getMemOffset()).toString(), L7);
		
		
		//now just let the ExprSTO know that it has a runtime addr, which is in L7
		res.setRuntimeAddr(new StorageLocation(L7));
		//this is going to make it store that runtime addr somewhere in temp mem
		//and then use that addr to to load and store the res' value from there, instead of from some bogus location
		
		
		//and load from there
		if (res.getParentType().isFloat()){
			AssemblyWriter.writeSmtAndComment ("now load from [L7] into F0", LOAD, L7, F0);
			res.setReg(F0);
		}else{
			AssemblyWriter.writeSmtAndComment ("now load from [L7] into L0", LOAD, L7, L0);
			res.setReg(L0);
		}	
		
		
		return res;
	}


}