package cs131w.symtab.ops.binary;

//---------------------------------------------------------------------
// relation op, < <= >= >
//---------------------------------------------------------------------
import cs131w.symtab.ops.Operator;
import  cs131w.lexer.*;
import  cs131w.parser.*;
import  cs131w.symtab.sto.*;
import  cs131w.symtab.types.*;
import  cs131w.error.*;
import  cs131w.assembly.*;


public abstract class RelOp extends BinaryOp{
	
	//----------------------------------------------------------------
	// ctor
	//----------------------------------------------------------------
	public 
	RelOp ()
	{
		super("", null);
	}
	

	public 
	RelOp (String strName, Token tok)
	{
		super(strName, tok);
	}
	
	//this function needs to check both operands if they are equal of coercible and return appr type
	public STO checkOperands(STO a, STO b) { 
		MyParser.debug(this + ": checking " + a + ", and " + b);
		
		if ( !a.getType().isNumeric()){
			
			return (new ErrorSTO( Formatter.toString(ErrorMsg.error1n_Expr, a.getType(), this)));
			
		}else if ( !b.getType().isNumeric()) { 
		
			return (new ErrorSTO( Formatter.toString(ErrorMsg.error1n_Expr, b.getType(), this)));
			
		} 	
		
		return a;
		
	}
	
	
	public STO 
	operate(STO a, STO b){
		MyParser.debug(this + ".operate: '" +a+ "', " + (b));
		STO res = new ExprSTO("relop", new BoolType());
		
		res.setRVal();
		return res;
	}
	
	//this needs to be overwritten by each binary op
	//here we just do the neccasary code and labels for a comp
	public STO
	AsmblyDoOp(STO res, STO a, STO b){
		
		MyParser.asmblyWriteVarSmt(COMMENT, "--- Begin Relational Op ---");
		
		String trueLabel = AssemblyWriter.getNextCmpTempLabel() + ".true";
		String endLabel = AssemblyWriter.getNextCmpTempLabel() + ".end";
		
		boolean b_changed = false;
		boolean a_changed = false;
		
		//both are already loaded, do the comp
		
		//first check if we need to promote a float
		if (b.getType().isFloat() && a.getType().isInt()){
	
			//promote src to float by using the cast op
			a = MyParser.DoBinaryExpr(new EmptySTO(new FloatType()), new CastOp(), a);
			//and load back up
			a.AsmblyLoad();
			//WARNING: now we are also responsible for freeing it from reg after using since we have changed the instance of src
			//and the calling AsmblyOp in Operator can't free this new instance, since it has no way of knowing
			//it will still try to clrReg the old instance, so we do it ourselves at the end
			a_changed = true;
		}

		//first check if we need to promote b float
		if (a.getType().isFloat() && b.getType().isInt()){
	
			//promote src to float by using the cast op
			b = MyParser.DoBinaryExpr(new EmptySTO(new FloatType()), new CastOp(), b);
			//and load back up
			b.AsmblyLoad();
			//WARNING: now we are also responsible for freeing it from reg after using since we have changed the instance of src
			//and the calling AsmblyOp in Operator can't free this new instance, since it has no way of knowing
			//it will still try to clrReg the old instance, so we do it ourselves at the end
			b_changed = true;
		}
		
		String op = CMP;
		String br = m_Template;
		
		//this needs to be written thru MyParser, coz it could be in global INIT
		if (a.getType().isFloat() || b.getType().isFloat()){
			op = FCMP;
			br = m_FloatTemplate;
		}
		
		MyParser.asmblyWriteVarSmt(op, a.getReg(), b.getReg());
		
		//now branch if true
		MyParser.asmblyWriteVarSmt(br, trueLabel);
		MyParser.asmblyWriteVarSmt(NOP);
		
		//write the false
		MyParser.asmblyWriteVarSmt(SET, "0", L0);
		//branch to end
		MyParser.asmblyWriteVarSmt(BA, endLabel);
		MyParser.asmblyWriteVarSmt(NOP);
		
		//true label
		MyParser.asmblyWriteVarSmt(LABEL, trueLabel);
		MyParser.asmblyWriteVarSmt(SET, "1", L0);
		
		//end
		MyParser.asmblyWriteVarSmt(LABEL, endLabel);
		
		//set that res is in l0
		res.setRegister(L0);
		
		if (b_changed) b.AsmblyClearReg();
		if (a_changed) a.AsmblyClearReg();
		
		MyParser.asmblyWriteVarSmt (COMMENT, "--- End Relational Op ---");
		//and return the res expr STO
		return res;
	
	}
	
	
	
	
}