package cs131w.symtab.ops.unary;

//---------------------------------------------------------------------
// basic math ops, like +-*/
//---------------------------------------------------------------------
import cs131w.symtab.ops.Operator;
import  cs131w.lexer.*;
import  cs131w.parser.*;
import  cs131w.symtab.sto.*;
import  cs131w.symtab.types.*;
import  cs131w.error.*;
import  cs131w.assembly.*;


public class StarOp extends UnaryOp{
	
	//----------------------------------------------------------------
	// ctor
	//----------------------------------------------------------------
	public 
	StarOp ()
	{
		super("*", new Token (sym.T_STAR, "*"));
	}
		
	//function that checks operand of this op
	public STO 
		checkOperand(STO sto){

			MyParser.debug(this + ": checking " + sto + ", isModifiable: " + sto.isModLValue());
			//we have to check that this is indeed a pointer or an array name
			if (!sto.getParentType().isBasicPtr()){	// && !sto.getParentType().isArray()){
				return new ErrorSTO(Formatter.toString(ErrorMsg.error15_Receiver, sto.getType()));
			}
	
			return sto;
			
		}
	
	@Override
	public STO operate(STO sto){
		//yes it is, so return a new ExprSTO with this pointer deref'fed once
		MyParser.debug(this+".operate: '" +sto+ "'.getParent= " + (sto.getParentType()));
		
		BasicPtrType ptr;
		Type res;
		STO ret;
		
		if (sto.getParentType().isBasicPtr()){
			//just do explicit cast call
			ptr = (BasicPtrType) sto.getParentType().getRootType();
			ret = new ExprSTO(ptr.derefOnce());
		}else{
			//in case of anything else, just return the underlying type
			ret = new ExprSTO(sto.getParentType().getUnderlyingType());
		}
		
		ret.setModLVal();
		MyParser.debug(this+".operate returning '" +ret+ "'");
		
		return ret;
		
	}
	
	
	//this needs to be overwritten by each unary op
	public STO
	AsmblyDoOp(STO res, STO a){
		//val is already loaded
		//have to do a quick check if its NULL
		
		//check for NULL
		AsmblyCheckNULL(res, a);
		//and check for dealloc SP 
		AsmblyCheckDeallocSP(res, a);
		
		
		String tar = L0;
		//so we just need to load the val from that ptr
		
		if (res.getType().isFloat())
			tar = F0;
		
		AssemblyWriter.writeSmtAndComment ("now load from "+a.getReg()+" into " + tar, LOAD, a.getReg(), tar);
		
		//signal that the result, ie. val that the pointer pointed to is now in L0 or F0
		res.setReg(tar);
		
		//and let the ExprSTO know that it has a runtime addr, which is in a.getReg()
		//for the case that we assigning
		res.setRuntimeAddr(new StorageLocation(a.getReg()));
		
		//and return
		return res;
		
	}
	
	private void
	AsmblyCheckNULL(STO res, STO a){
		
		AssemblyWriter.writeSmt (COMMENT, "--- Begin NULL Ptr deref check ---");
		String endLabel = AssemblyWriter.getNextCmpTempLabel() + "_NullCh.end";
		
		//a is already loaded, do the comp
		AssemblyWriter.writeSmtAndComment("compare to 0 to test if NULL", CMP, a.getReg(), G0);
		
		//now branch if not true
		AssemblyWriter.writeSmt(BNE, endLabel);
		AssemblyWriter.writeSmt(NOP);
		
		//output deref NULL ptr msg
		AssemblyWriter.writeSmtAndComment("set the 'NULL ptr' msg into O0", SET, "NULLPtr", O0);
		AssemblyWriter.writeSmtAndComment("output", CALL, "printf");
		AssemblyWriter.writeSmt(NOP);
		//and call exit
		AssemblyWriter.writeSmtAndComment("call exit to set return addr to OS", CALL, "exit");
		AssemblyWriter.writeSmt(NOP);
		AssemblyWriter.writeSmtAndComment("exit(1)", SET, "1", O0);
		
		AssemblyWriter.writeSmtAndComment("and ret...", RET);
		AssemblyWriter.writeSmtAndComment("return to OS", RESTORE);
		
		
		//end
		AssemblyWriter.writeSmt(LABEL, endLabel);
		AssemblyWriter.writeSmt(COMMENT, "--- End NULL Ptr deref check ---");
		
		
	}


	private void
	AsmblyCheckDeallocSP(STO res, STO a){
		
		AssemblyWriter.writeSmt(COMMENT, "--- Begin Dealloc Stack Ptr deref check ---");
		
		//String failLabel = AssemblyWriter.getNextCmpTempLabel() + "_deAllocCh.fail";
		String endLabel = AssemblyWriter.getNextCmpTempLabel() + "_deAlloc.end";
		
		//check if lower than lowest SP ever reached, than its ok
		//AssemblyWriter.writeSmtAndComment("set lowest offs of SP into L7", SUB, SP, new Integer(MyParser.getFrameOffset()).toString(), L7);
		
		//set label
		AssemblyWriter.writeSmtAndComment ("set label of global lowest SP val into G1", SET, "GLOB.LowSP", G1);
		//and load into G2
		AssemblyWriter.writeSmtAndComment ("load", LOAD, G1, G2);
		AssemblyWriter.writeSmtAndComment("compare to lowest SP", CMP, a.getReg(), G2);
		
		//now branch if lower to end label
		AssemblyWriter.writeSmt(BLU, endLabel);
		AssemblyWriter.writeSmt(NOP);
		
		//check if higher than SP+92
		AssemblyWriter.writeSmtAndComment("set SP+92 into L7", ADD, SP, "92", L7);
		AssemblyWriter.writeSmtAndComment("check if higher than SP+92", CMP, a.getReg(), L7);
		
		//now branch if higher to end label
		AssemblyWriter.writeSmt(BGU, endLabel);
		AssemblyWriter.writeSmt(NOP);
		
		
		//and fail
		//AssemblyWriter.writeSmt(LABEL, failLabel);
		
		//output dealloc ptr msg
		AssemblyWriter.writeSmtAndComment("set 'Deallocated stack ptr' msg into O0", SET, "DeAllPtr", O0);
		AssemblyWriter.writeSmtAndComment("output", CALL, "printf");
		AssemblyWriter.writeSmt(NOP);
		//and call exit
		AssemblyWriter.writeSmtAndComment("call exit to set return addr to OS", CALL, "exit");
		AssemblyWriter.writeSmt(NOP);
		AssemblyWriter.writeSmtAndComment("exit(1)", SET, "1", O0);
		
		AssemblyWriter.writeSmtAndComment("and ret...", RET);
		AssemblyWriter.writeSmtAndComment("return to OS", RESTORE);
		
		
		//end
		AssemblyWriter.writeSmt(LABEL, endLabel);
		AssemblyWriter.writeSmt(COMMENT, "--- End Dealloc Stack Ptr deref check ---");
		
		
	}
	
	

}