package cs131w.symtab.ops.unary;

//---------------------------------------------------------------------
// basic math ops, like +-*/
//---------------------------------------------------------------------
import cs131w.symtab.ops.Operator;
import  cs131w.lexer.*;
import  cs131w.parser.*;
import  cs131w.symtab.sto.*;
import  cs131w.symtab.types.*;
import  cs131w.error.*;


public abstract class UnaryIncDecOp extends UnaryOp{
	private boolean 	isPostFix;
	protected String	m_Template_am;
	//----------------------------------------------------------------
	// ctor
	//----------------------------------------------------------------
	public 
	UnaryIncDecOp ()
	{
		super("", null);
	}
	
	public 
	UnaryIncDecOp (String strName, Token tok)
	{
		super(strName, tok);
	}
	
	//protected boolean isModifyingOp(){ return true; }

	
	//function that checks both operands of this op
	public STO 
		checkOperand(STO a){
			MyParser.debug(this + ": checking " + a);
			
			if (!a.getType().isPtr() && !a.getType().isNumeric()){
				return (new ErrorSTO( Formatter.toString(ErrorMsg.error2_Type, a.getType(), this)));
			} else if (!a.isModLValue()) {
				return (new ErrorSTO( Formatter.toString(ErrorMsg.error2_Lval, this)));
			} 	
			
			return a;
		}
	
	@Override
	public STO operate(STO a){
		MyParser.debug(this + ".operate: '" + a);
		return new ExprSTO(a.getName() + "INC/DEC", a.getType());
	}

	public STO
	AsmblyDoOp(STO res, STO a){
		if (a.getType().isFloat())
			return AsmblyDoFlOp(res, a);
		else
			return AsmblyDoIntOp(res, a);
	}

	//this needs to be overwritten by each unary op
	//Int version
	public STO
	AsmblyDoIntOp(STO res, STO a){
		//a is the varSTO and is always inc/dec'd
		//res is the exprSTO and should only be incremented in the prefix case.
		//no we do the actual op in asmbly
		//TODO should just use L0 rather than loading res, haven't bothered yet
		//this needs to be written thru MyParser, coz it could be in global INIT
		if(this.isPostFix()){
			MyParser.asmblyWriteVarSmtComm (this+": " + a.getReg(), MOVE, a.getReg(), L0);
			//MyParser.asmblyWriteVarSmtComm (this+": Move into L0: " + a.getReg(), ADD, a.getReg(), "0", L0);
			
		}
		
		//do the op now
		if (a.getType().isNumeric()){
		
			MyParser.asmblyWriteVarSmtComm (this+": Do the inc: " + a.getReg(), m_Template, a.getReg());
		
		}else if(a.getType().isPtr()){
			//if its a ptr, we need to INC/DEC by size of underlying type, rather than just 1
			int size = a.getType().getPointedToType().getSize();
			MyParser.asmblyWriteVarSmtComm (this+": (Ptr) do inc: " + a.getReg(), m_Template_am, new Integer(size).toString(), a.getReg());
		}
		
		if(!this.isPostFix()){
			MyParser.asmblyWriteVarSmtComm (this+": " + a.getReg(), MOVE, a.getReg(), L0);
			//MyParser.asmblyWriteVarSmtComm (this+": Move into L0: " + a.getReg(), ADD, a.getReg(), "0", L0);
		}
		
		res.setReg(L0);
		
		//and return the res expr STO
		return res;
	}
	
	//this needs to be overwritten by each unary op
	//Int version
	public STO
	AsmblyDoFlOp(STO res, STO a){

		if(this.isPostFix()){
			MyParser.asmblyWriteVarSmtComm (this+": " + a.getReg(), FMOVE, a.getReg(), F0);
		}
		
		//do the op now, since this is float, we need to add 1.0
		//get global 1.0
		MyParser.asmblyWriteVarSmtComm ("get addr of global 1.0 const", SET, "GLOB.1_0", L0);
		MyParser.asmblyWriteVarSmtComm ("and load into F31", LOAD, L0, F31);
		
		
		MyParser.asmblyWriteVarSmtComm (this+": Do the inc/dec in float: " + a.getReg(), m_FloatTemplate, a.getReg(), F31, a.getReg());
		
		if(!this.isPostFix()){
			MyParser.asmblyWriteVarSmtComm (this+": " + a.getReg(), FMOVE, a.getReg(), F0);
		}
		
		res.setReg(F0);
		
		//and return the res ExprSTO
		return res;
	}
	

	public boolean isPostFix() {
		return isPostFix;
	}

	public void setPostFix() {
		isPostFix = true;
	}
	
	public void setPreFix() {
		isPostFix = false;
	} 
	

}