package cs131w.symtab.sto;
//---------------------------------------------------------------------
//
//---------------------------------------------------------------------
import  cs131w.symtab.types.*;
import  cs131w.parser.MyParser;
import  cs131w.error.*;
import  cs131w.assembly.*;
import  cs131w.symtab.*;

public class AdressableSTO extends STO
{

	public 
	AdressableSTO (String strName, Type typ)
	{
		super(strName, typ);
	}	
	
	
	public String
	toString()
	{
		if (MyParser.DEV){
			String ad = "";
			if (isInReg())
				ad = m_register;
			else
				ad = getAddrStr();
				
			return ("(AddrSto)" + m_strName + ":" + m_type + "@" + ad);
		}else
			return (m_strName);
		
	}

	//----------------------------------------------------------------
	//
	//----------------------------------------------------------------
	
	//this func creates this var or value if its not already created
	//takes care of using the right code method
	public void
	AsmblyCreate(){ 
		if (isError() || mem_base == null) return;
		MyParser.debug(this + ".AsmblyCreate called. SHOULD NOT HAPPEN!");
		asmbl_created = true;
	}
	


	//loads this STO into a reg, if its in mem, doesnt care where
	public void
	AsmblyLoad(){ 
		MyParser.debug(this + ".AsmblyLoad() called in Addresable.");
		
		if (isInReg()) return;
		String in = "";
		
		if (getParentType().isFloat())
			in = Regs.getNextFloatReg();
		else
		 	in = Regs.getNextReg();
		
		AsmblyLoad(in);
	}
	
	//loads this STO into a reg, if its in mem, do care about where it is stored into
	//also this assumes that this reg is free, does not check with Regs, or tells it to reserve
	public void
	AsmblyLoad(String into){ 
		MyParser.debug(this + ".AsmblyLoadInto("+into+") called in Addresable.");
		if (isError()) return;
		
		if (!asmbl_created) AsmblyCreate();
		
		//well anything that extends this class is addressable - ie in temp somewhere
		//just believe it w/out testing, even if it might not be _Addressable_ in realiter 
		//if ((!isExpr() && !m_isAddressable)) return;
		
		
		//if already loaded: return
		if (isInReg() && m_register == into) return;
		
		//if loaded but somewhere else, move it
		if (isInReg() && m_register != into){
		 	//need to move it there
			
			String op = MOVE;

			//this needs to be written thru MyParser, coz it could be in global INIT
			if (getType().isFloat())
				op = FMOVE;
			
			MyParser.asmblyWriteVarSmt(op, m_register, into);
			//and free old one if it happens to be local
			Regs.FreeReg(m_register);
			m_register = into;
			return;
		}
	
		//first check if it has a runtime addr, if so we need to load the address from there
		if (hasRuntimeAddr()){
			loadRuntimeAddrInto(L7);
			//and now do the actual load from that addr
			MyParser.asmblyWriteVarSmtComm ("and load into reg " + into, LOAD, L7, into);
			mem_runtimeAddr.clrReg();
			
		}else{
			//no runtime address, check if it has base
			if (mem_base == null) return;
			
			//load into mem
			if (mem_isLocal){
				//AssemblyWriter.writeSmt(LOAD, getAddrStr(), into);
				MyParser.asmblyWriteVarSmtComm ("Load '" + this + "' into " + into, LOAD, getAddrStr(), into);
				//now mark as in reg
			}else{
		
				MyParser.asmblyWriteVarSmtComm ("set addr of '" + this + "' into L7", SET, getMemBase(), L7);
				//and now do the loading
				MyParser.asmblyWriteVarSmtComm ("and load into reg " + into, LOAD, L7, into);
			}
		}
		
		m_register = into;
	}

	public void
	AsmblyLoadAddr(){
		MyParser.debug(this + ".AsmblyLoadAddr() called.");
		AsmblyLoadAddr(Regs.getNextReg());
	}
	
	public void
	AsmblyLoadAddr(String into){
		MyParser.debug(this + ".AsmblyLoadAddr("+into+") called.");
		if (isError()) return;
		if (!asmbl_created) AsmblyCreate();
	
		if (hasRuntimeAddr()){
			loadRuntimeAddrInto(into);
		}else{
			
			//no runtime address, check if it has base
			if (mem_base == null) return;
			
			//now load the addr
			if (mem_isLocal){
			
				if (mem_offset > 0)
					MyParser.asmblyWriteVarSmtComm ("Load addr of '" + this + "' into " + into, ADD, getMemBase(), new Integer(mem_offset).toString(), into);
				else
					MyParser.asmblyWriteVarSmtComm ("Load addr of '" + this + "' into " + into, SUB, getMemBase(), new Integer(Math.abs(mem_offset)).toString(), into);
	
			}else{
				MyParser.asmblyWriteVarSmtComm ("set addr of '" + this + "' into " + into, SET, getMemBase(), into);
			}
		}
	} 
	
	
	//stores this STO back from reg into mem where it was created, and thus frees this reg
	public void
	AsmblyStore(){ 
		MyParser.debug(this + ".AsmblyStore called.");

		//do not store if not addressable or an expression - which lives in temp space (so it is addressable), or if it is an error
		if (isError()) return;
		if (!asmbl_created) AsmblyCreate();
		
		if (hasRuntimeAddr()){
			loadRuntimeAddrInto(L7);
			//and now do the actual store to that addr
			AsmblyStore(L7);
			//and tell the runtime addr that it is not in L7 anymore, since L7 gets overwritten by each op
			mem_runtimeAddr.clrReg();
			
		}else{
		
			//no runtime address, check if it has base
			if (mem_base == null) return;
			
			//store away
			if (mem_isLocal){
				AsmblyStore(getAddrStr());
			}else{
				MyParser.asmblyWriteVarSmtComm ("set addr of '" + this + "' into L7", SET, getMemBase(), L7);
				AsmblyStore(L7);
			}
		}
	}
	
	
	//stores this STO back to specified addr in to (which is either a name of a reg that contains address or addr. string like %fp-68)
	public void
	AsmblyStore(String to){ 
		MyParser.debug(this + ".AsmblyStore("+to+") called.");

		if (isError()) return;
		
		//if not loaded: return
		if (!isInReg()) {
			MyParser.asmblyWriteVarSmtComm ("Can't store " + this + "! " + getName() + " is not in reg :(", "");
			return;
		}
		
		MyParser.asmblyWriteVarSmtComm ("Store '" + this + "' from "+m_register+" to " + to, STORE, m_register, to);
		
		Regs.FreeReg(m_register);
		m_register = null;
		
	}
	
	

	public void
	AsmblyCOut(){ 
		MyParser.debug(this + ".AsmblyCOut called.");
		
		//AsmblyLoad();
		
		
		//first decide what kind of type we are printing
		//if int ptr, ie. array or struct as well, just print Int val
		if (getParentType().isInt()){
			
			AssemblyWriter.writeSmt(SET, "intFmt", O0);
			AsmblyLoad(O1);
			AssemblyWriter.writeSmt(CALL, "printf");
			AssemblyWriter.writeSmt(NOP);
			
			
		}else if (getParentType().isPtr()){
			
			AssemblyWriter.writeSmt(SET, "PtrT", O0);
			AsmblyLoad(O1);
			AssemblyWriter.writeSmt(CALL, "printf");
			AssemblyWriter.writeSmt(NOP);

		}else if (getParentType().isArray()){
			
			AssemblyWriter.writeSmt(SET, "ArrayT", O0);
			AsmblyLoad(O1);
			AssemblyWriter.writeSmt(CALL, "printf");
			AssemblyWriter.writeSmt(NOP);
			
		}else if (getParentType().isStruct()){
			
			AssemblyWriter.writeSmt(SET, "StructT", O0);
			AsmblyLoad(O1);
			AssemblyWriter.writeSmt(CALL, "printf");
			AssemblyWriter.writeSmt(NOP);
		
		}else if (getParentType().isFloat()){
			
			AsmblyLoad(F0);
			AssemblyWriter.writeSmt(CALL, "printFloat");
			AssemblyWriter.writeSmt(NOP);
			
		}else if (getParentType().isBool()){
			//now branch depending wether its true or false
			String false_label = AssemblyWriter.getNextIfTempLabel();
			String end_label = AssemblyWriter.getNextIfTempLabel();
			
			AsmblyLoad(L1);
			AssemblyWriter.writeSmt(CMP, L1, G0);
			
			AssemblyWriter.writeSmt(BE, false_label);
			AssemblyWriter.writeSmt(NOP);
			AssemblyWriter.writeSmt(SET, "boolT", O0);
			AssemblyWriter.writeSmt(BA, end_label);
			AssemblyWriter.writeSmt(NOP);
			
			AssemblyWriter.writeSmt(LABEL, false_label);
			AssemblyWriter.writeSmt(SET, "boolF", O0);
			AssemblyWriter.writeSmt(LABEL, end_label);
			
			//and now call printf
			AssemblyWriter.writeSmt(CALL, "printf");
			AssemblyWriter.writeSmt(NOP);

		}else{
			//unknown type
			MyParser.debug(this + ".COut: unknown type! FIX ME!");
		}
		
		
		AsmblyClearReg();
	}

	
}
