package cs131w.symtab.sto;
//---------------------------------------------------------------------
//
//---------------------------------------------------------------------
import  cs131w.symtab.types.*;
import  cs131w.symtab.*;
import  cs131w.symtab.ops.*;
import  cs131w.symtab.ops.binary.*;
import  cs131w.error.*;
import  cs131w.assembly.*;
import  cs131w.parser.MyParser;

public class ConstSTO extends AdressableSTO
{

    private Double		m_value;
    private String		m_StrValue;

	//----------------------------------------------------------------
	// basic ctors
	//----------------------------------------------------------------
	public 
	ConstSTO ()
	{
		this ("", new VoidType(), new Double(0));
	}

	//ctor for a constant int value
	public 
	ConstSTO (int val)
	{
		this (new Integer(val).toString(), new IntType());
		setValue(val);
	}

	//ctor for a constant string
	public 
	ConstSTO (String s)
	{	//set name, underlying type and s as the val
		this (s, new StringType(s), s);
	}

	/*	
		public 
		ConstSTO (String strName)
		{
			this (strName, new VoidType(), new Double(0));
		}
	*/

	//ctor for a constant int value
	public 
	ConstSTO (double val)
	{
		this (new Double(val).toString(), new IntType());
		setValue(val);
	}


	//ctor for a constant float value
	public 
	ConstSTO (float val)
	{
		this (new Float(val).toString(), new FloatType());
		setValue(val);
	}

	//ctor for a constant bool value
	public 
	ConstSTO (boolean val)
	{
		this (new Boolean(val).toString(), new BoolType());
		m_value = new Double((val) ? 1 : 0);
	}
	
	public 
	ConstSTO (String strName, Type t)
	{
		this (strName, t, new Double(0));
	}


	//----------------------------------------------------------------
	//ctor with val
	//----------------------------------------------------------------
	public 
	ConstSTO (String strName, Type t, Double val)
	{
		super (strName, t);
		m_value = val;
		m_StrValue = "";
       	//set this to an RVal, createors have to change if desired
		setRVal();
		m_label = "const." + m_strName + "$L" + MyParser.GetLineNum();
		
		MyParser.debug(this + ".ctor returning.");
		
	}

	public 
	ConstSTO (String strName, Type t, String val)
	{
		super (strName, t);
		m_value = new Double(0);
		m_StrValue = val;
       	//set this to an RVal, createors have to change if desired
		setRVal();
		MyParser.debug(this + ".ctor returning.");
		
	}


	//----------------------------------------------------------------
	//
	//----------------------------------------------------------------
	public boolean
	isConst () 
	{
		return true;
	}

	public boolean	isNull ()
	{ 
		return (m_value.isNaN()); 
	}


	public String
	toString()
	{
		if (MyParser.DEV){
			String t = m_value.toString();
			if (m_StrValue != "") t = "'" + m_StrValue + "'";
			String ad = "";
			if (isInReg())
				ad = m_register;
			else
				ad = getAddrStr();
				
			
			return ("(Const)" + m_strName + ":" + m_type + "=" + t + "@" + ad);
		}else
			return (m_strName);
	}	

	//----------------------------------------------------------------
	//
	//----------------------------------------------------------------
	public void
	setValue (double val) 
	{
		m_value = new Double(val);
	}

	public Double
	getValue () 
	{
		return m_value;
	}

	public int
	getIntValue () 
	{
		return m_value.intValue();
	}

	public float
	getFloatValue () 
	{
		return m_value.floatValue();
	}

	public boolean
	getBoolValue () 
	{
		return (m_value.intValue() != 0);
	}

	public String
	getValueAsString(){
		
		String ret = "";
		
		if (this.getParentType().isInt()){
			ret = new Integer(getIntValue()).toString();
		}else if (this.getParentType().isFloat()){
			ret =  new Float(getFloatValue()).toString();
		}else if (this.getParentType().isBool()){
			ret = new Integer(getIntValue()).toString();
		}else if (this.getParentType().isPtr() && m_value.isNaN()){
			ret = "0";
		}else if(this.getParentType().isString())
			ret = m_StrValue;
			
		return ret;
	}
	
	
	public String
	getStringValue () 
	{
		return m_StrValue;
	}

	public void
	setStringValue (String s) 
	{
		m_StrValue = s;
	}

	@Override
	public Object 
	clone() { 
		ConstSTO ret = new ConstSTO(new String(m_strName), (Type)m_type.clone(), new Double(m_value));
		ret.setStringValue(m_StrValue);
		ret.setIsModifiable(m_isModifiable);
		ret.setIsAddressable(m_isAddressable);
		
		//asmbly stuff
		ret.setReg(m_register);
		ret.setMemBase(mem_base);
		ret.setMemOffset(mem_offset);
		ret.setIsLocal(mem_isLocal);
		if (m_static) ret.setStatic();
		if (asmbl_created) ret.setAsmblyCreated();
		if (hasRuntimeAddr()) ret.setRuntimeAddr((StorageLocation)mem_runtimeAddr.clone());
		
		return ret;
	}

	//----------------------------------------------------------------
	// Factory method to do the checking and make a new ConstSTO for us
	//----------------------------------------------------------------

	public static STO
	create (STO element, Type t){
		
		SymbolTable st = MyParser.getSymTab();
		

		//verify that what we're assigning to the constant is constant or can be folded to one
		if (element.isError()){
			return new ErrorSTO  (element, Formatter.toString(ErrorMsg.error8_CompileTime, element.getName()));
		}	

		if (st.accessLocal (element.getName()) != null)
		{
			return MyParser.STO_getRedeclIDReport (element.getName());
		}


		if(!element.isConst() && !element.isFunc()){
			return new ErrorSTO  (element, Formatter.toString(ErrorMsg.error8_CompileTime, element.getName()));
		}

		ConstSTO target;

		if (element.isFunc()){
			target = new ConstSTO(element.getName(), t, new Double(0));
		}else{

			ConstSTO sto = (ConstSTO) element;
			target = new ConstSTO(element.getName(), t, sto.getValue());

		}

		//check assignability by using AssignOp()

		Operator o = new AssignOp();
		//set to mod LVal so the check goes through
		target.setModLVal();

		STO res = o.checkOperands(target, element);
		if (res.isError()) {
				return new ErrorSTO (element, Formatter.toString(ErrorMsg.error8_Assign, element.getType(), t));
		}

		//and set back, to not mod but addressable
		target.setIsModifiable(false);
		target.setIsAddressable(true);
		return target;
		
	}
	
	
	
	
	//----------------------------------------------------------------
	//	Assembly functions
	//----------------------------------------------------------------
	
	public void
	AsmblyCreate(){ 
		if (asmbl_created) return;
		
		MyParser.debug(this + ".AsmblyCreate called.");
		
		//There are 4 possible options for this const:
		//either its a var that is a const, meaning its addressable
		//can bei either 
		//		-local:		this just creates the value and stores is somewhere in mem
		//		-global:	this sets a global init'ed label
		//or its a temp value, that is a literal (which is an RVal)
		//		-local:		just create the value somewhere in mem 
		//		-global		this also creates it somewhere in mem, the same, 
		//					but needs to be done in global.INIT which is taken care of by MyParser
		//if its addressable that means its a var
		//or its not addresable, which means its just a local const
		
		asmbl_created = true;
		
		//since for global var inits we create a temp function this is also local
		//and the func and AsmblyWriter takes care of inserting it into the right location
		//so we just emit a global var init if this is both addrsable and global
		if (getIsAddressable() && MyParser.isInGlobalSpace()){
			AsmblyCreateVarGlobal();
		}else{
			AsmblyCreateTemp();
		}
		
	}

	//this creates and stores a temp value in mem somewhere
	public void
	AsmblyCreateTemp(){ 
		MyParser.debug(this + ".AsmblyCreate: ctoring temp.");
		int off = MyParser.increaseFrameOffset(getSize());
		markLocal(off);
		
		MyParser.asmblyWriteVarSmt (COMMENT, "creating local const '"+this+"' at offset %fp" + off);
		
		//if this is float we need to store the float val in a label
		if (getParentType().isFloat() || getParentType().isString()){
			boolean fl = getParentType().isFloat();
			
			String lbl = ( (fl) ? AssemblyWriter.getNextTempLabel() : AssemblyWriter.getNextStrTempLabel());
			AssemblyWriter.switchSection(DATA);
			//AssemblyWriter.writeSmt(LABEL, lbl);
			AssemblyWriter.writeSmt( (fl) ? FLOAT_VAR_INIT : STRING_VAR_INIT, lbl, getValueAsString() );
			
			//and remember in this sto
			mem_base = lbl;
			mem_offset = 0;
			mem_isLocal = false;
			AssemblyWriter.switchSection(TEXT);
			return;
		}
		
		//else this means its a bool or int
		//we just set it somewhere and store it in temp
		//make sure to call the MyParser func here, which either sets it into current place or global.INIT if we are in global space
		MyParser.asmblyWriteVarSmtComm ("set imm. val into L6", SET, getValueAsString(), L6);
		
		//and mark that we are set in L7 for now
		setReg(L6);
		//and store back into mem
		AsmblyStore();
		
		//MyParser.asmblyWriteVarSmtComm ("and store into " + , STORE, L7, new Integer(off).toString() );
		
	}
	
	//this functions emits code for making this a global constant
	private void
	AsmblyCreateVarGlobal(){
		MyParser.debug(this + ".AsmblyCreate: ctoring global var.");
		markGlobal();
		
		//switch to rodata section
		MyParser.debug(this + ".emitGlobalAssembly switching to rodata");
		AssemblyWriter.switchSection(RODATA);
		
		if (!m_static){
			//write global marker
			MyParser.debug(this + ".emitGlobalAssembly writing global marker");
			AssemblyWriter.writeSmt(GLOBAL, this.getLabel());
		}
				
		
		if (getParentType().isFloat()){

			MyParser.debug(this + ".emitGlobalAssembly writing var as float:" + getValueAsString());
			AssemblyWriter.writeSmt(FLOAT_VAR_INIT, this.getLabel(), getValueAsString() );

		//now see what kind of var it is
		}else{
			//write the var
			MyParser.debug(this + ".emitGlobalAssembly writing var as int:" + getValueAsString());
			AssemblyWriter.writeSmt(WORD_VAR_INIT, this.getLabel(), getValueAsString() );
		}
			
		MyParser.debug(this + ".emitGlobalAssembly returning");
		AssemblyWriter.switchSection(TEXT);
		
		
	}
	
	public void
	AsmblyCOut(){ 
		MyParser.debug(this + ".AsmblyCOut called.");
		if (!asmbl_created) AsmblyCreate();
		//first load this puppy up
		
		//first decide what kind of type we are printing
		if (this.getParentType().isInt()){
			AssemblyWriter.writeSmt(SET, "intFmt", O0);
			AsmblyLoad(O1);
			AssemblyWriter.writeSmt(CALL, "printf");
			AssemblyWriter.writeSmt(NOP);
			
			
		}else if (this.getParentType().isFloat()){
			AsmblyLoad(F0);
			AssemblyWriter.writeSmt(CALL, "printFloat");
			AssemblyWriter.writeSmt(NOP);
			
		}else if (this.getParentType().isBool()){
			//see if its true or false
			if (getBoolValue()){
				AssemblyWriter.writeSmt(SET, "boolT", O0);
			}else{
				AssemblyWriter.writeSmt(SET, "boolF", O0);
			}
			
			AssemblyWriter.writeSmt(CALL, "printf");
			AssemblyWriter.writeSmt(NOP);

		}else if (this.getParentType().isString()){
			//AssemblyWriter.writeSmt(SET, "strFmt", O0);
			//just set the string's addr and call printf
			
			AssemblyWriter.writeSmt(SET, mem_base, O0);
			AssemblyWriter.writeSmt(CALL, "printf");
			AssemblyWriter.writeSmt(NOP);

		}
	}
	

}
