package cs131w.symtab.sto;
//---------------------------------------------------------------------
//
//---------------------------------------------------------------------
import  cs131w.symtab.types.*;
import  cs131w.error.*;
import  cs131w.symtab.*;
import  cs131w.assembly.*;
import  cs131w.symtab.ops.binary.*;
import  cs131w.parser.MyParser;
import 	java.util.Vector;

public class FuncSTO extends STO
{
		protected boolean	m_bReturnFound = false;
		protected int		m_scopeDeepness;
		
		//asmbly vars
		//offset of %fp to create a new var and temp values at, also remembers number needed for reserving at the beginning
		protected int		m_FrameOffset = 0;
		//scope to remember size of local vars
		protected Scope		m_localScope;
		protected String	m_parentName;
		
	//----------------------------------------------------------------
	//
	//----------------------------------------------------------------
	public 
	FuncSTO ()
	{
		this("func");
	}

	public 
	FuncSTO (String strName)
	{
		this(strName, new VoidType());
	}
	
	public 
	FuncSTO (String strName, Type t){
		this(strName, t, 0, "");
	}
	
	public
	FuncSTO(String strName, Type t, int deep, String parent)
	{
		//just set the type to the func proto type
		super (strName, t);
		setRVal();
		m_scopeDeepness = deep;
		m_parentName = parent;
		
		regenLabel();
	}

	public void
	regenLabel ()
	{
		
		if (m_parentName.equals("main")){
			m_label = m_parentName;
		
		//if sto is global (not local or not struct member), fall back to just the mangled name w/out LineNum
		}else if ( (!mem_isLocal || !m_isStructMember)){
			m_label = m_parentName + getSafeMangled();
		}else{
			m_label = "fun." + m_parentName + getSafeMangled() + "$L" + MyParser.GetLineNum();
		}
	}

	private String
	getSafeMangled(){
		String s = getMangledName().replace('[','_').replace(']','_').replace('*', '.');
		if (s.length() > 10)
			return AssemblyWriter.getNextFunTempLabel();
		else
			return s;
	}

	//----------------------------------------------------------------
	//
	//----------------------------------------------------------------
	public boolean
	isFunc () 
	{ 
		return true;
   	}


	//----------------------------------------------------------------
	//
	//----------------------------------------------------------------
	public int
	getScopeLevel(){
		return m_scopeDeepness;
	}


	//regardless of return type a funcSTO always has size 0!
	public int
	getSize(){
		return 0;
	}

	public Type
	getReturnType ()
	{
		return ((FuncProtoType)m_type).getReturnType();
	}

	public ParamSTO
	getParam (int i)
	{
		return ((FuncProtoType)m_type).getParam(i);
	}

	public FuncProtoType
	getProtoType ()
	{
		return ((FuncProtoType) m_type);
	}

	public Type
	getType ()
	{
		//return getReturnType();
		return getProtoType ();
	}

	//----------------------------------------------------------------

	public void
	setScope (Scope s)
	{
		m_localScope = s;
	}

	public Scope
	getScope ()
	{
		return m_localScope;
	}

	//----------------------------------------------------------------

	public int
	getFrameOffset ()
	{
		return m_FrameOffset;
	}

	public void
	setFrameOffset (int off)
	{
		m_FrameOffset = off;
	}
	
	//increases by spec. amount and returns old
	public int
	increaseFrameOffset ()
	{
		return increaseFrameOffset(4);
	}
	
	public int
	increaseFrameOffset (int by)
	{
		//int old = m_FrameOffset;
		m_FrameOffset -= by;
		return m_FrameOffset;
	}


	//----------------------------------------------------------------
	// If we have found a return already 
	//----------------------------------------------------------------
	public void
	ReturnFound ()
	{
		m_bReturnFound = true;
	}

	public boolean
	getReturnFound ()
	{
		return m_bReturnFound;
	}

	public void
	setReturnFound (boolean f)
	{
		m_bReturnFound = f;
	}


	public STO
	DoReturnStmt (STO ret){
		//we get called on ourselves as the func
		//STO func = m_symtab.getFunc();
		STO res = ret;
		
		//set that we have found a return stmt, if within top scope
		if (this.getScopeLevel() == MyParser.getSymTab().getLevel())
			this.ReturnFound();
		
		//no check that we are passing something meaningful if return type is not void
		if (ret == null && !this.getType().getReturnType().isVoid()){
			res = new ErrorSTO (ret, ErrorMsg.error6a_Return_expr);

		}else if (ret == null && this.getType().getReturnType().isVoid()){
			//everything good, just return
			res =  this;

		//and check if stmt is assignable to return type, if not void
		}else if (! ret.getType().isAssignable(this.getType().getReturnType())){
			res =  new ErrorSTO (ret, Formatter.toString(ErrorMsg.error6a_Return_type, ret.getType(), this.getType().getReturnType()));
		}
		
		return res;
	}



	//----------------------------------------------------------------
	//----------------------------------------------------------------
	public String
	toString()
	{
		if (MyParser.DEV)
			return ("(Func)" + m_strName + "():" + m_type);
		else
			return (m_strName);
		
	}	

	public Object 
	clone(){		// throws CloneNotSupportedException {
		
		FuncSTO ret = new FuncSTO(new String(m_strName), (Type)m_type.clone(), m_scopeDeepness, m_parentName);
		ret.setReturnFound(m_bReturnFound);
		
		ret.setIsModifiable(m_isModifiable);
		ret.setIsAddressable(m_isAddressable);

		ret.setFrameOffset(m_FrameOffset);
		ret.setScope(m_localScope);
		//asmbly
		ret.setReg(m_register);
		ret.setMemBase(mem_base);
		ret.setMemOffset(mem_offset);
		ret.setIsLocal(mem_isLocal);
		if (m_static) ret.setStatic();
		if (asmbl_created) ret.setAsmblyCreated();
		if (hasRuntimeAddr()) ret.setRuntimeAddr((StorageLocation)mem_runtimeAddr.clone());
		
		return ret;
	}
	
	
	public String getLabel(){
		//TODO is this any good? No: main will not have the name "main" anymore!
		return m_label;
		//return ".fun_" + getName() + getMangledName();
	}
	


	//----------------------------------------------------------------
	//	Assembly functions
	//	these emit the necc. assembly code for doing prescribed action
	//----------------------------------------------------------------
	//starts creation of this function in assemlby
	
	public void
	AsmblyCreate(){ 
		MyParser.debug(this + ".AsmblyCreate called.");
		
		
		//String name = m_parentName + getMangledName();
		//String name = m_label;
		
		//if we are not inside a struct, need to add it to global
		if (!MyParser.getSymTab().isDefiningStruct()){
			//write global marker
			AssemblyWriter.writeSmt(GLOBAL, m_label);
		}
		
		//set the label
		AssemblyWriter.writeSmt(LABEL, m_label);
		//set the save FP instruction into g1
		AssemblyWriter.writeSmt(SET, m_label + ".SAVE", G1);
		AssemblyWriter.writeSmt(SAVE, SP, G1, SP);
		
		//and save the lowest SP if smaller than current
		saveLowestSP();
		
		//now save all the params on the stack, starting at %fp+68
		int start = 68;
		
		int start_reg = 0;
		
		//if inside of struct, load this* into o0
		if (m_isStructMember){
			AssemblyWriter.writeSmt(STORE, "%i0", FP + "+" + start);
			start += 4;
			start_reg = 1;
		}
		
		//save params passed in %i to stack, only 1 -6
		for (int i = 0; i < getProtoType().getNumParams(); i++){
			
			
			ParamSTO arg = getParam(i);
			String b = "%i";
			 
			//if its float and not a ref, get from %f reg instead
			if (arg.getParentType().isFloat() && !arg.getIsRef()) 
				b = "%f";
			
			//only save from regs for params less than 6
			if (start_reg < 6)
				AssemblyWriter.writeSmt(STORE, b + start_reg, FP + "+" + start);
			
			//and remember the addr inside ParamSTO
			arg.markLocal(start);
			start += 4;
			start_reg++;
			
			
		}
	
		//and call global.INIT if we are in main()
		//MyParser.debug(this + ": checking global init: " + name);
		if (m_label.equals("main")){
			AssemblyWriter.addGlobalInitCall();
		}
		asmbl_created = true;
		
	}

	//finish this STO, ie. epilogue in this case
	public void
	AsmblyFinish(){ 
		MyParser.debug(this + ".AsmblyFinish called.");

		//write ret and restore only if there was no return stmt found at top level
		if (!getReturnFound()){
			AssemblyWriter.writeSmt(RET);
			AssemblyWriter.writeSmt(RESTORE);
		}
		
		//and set the bytes we need to save at the beginning into label
		AssemblyWriter.writeSmt(VAR, m_label + ".SAVE", "-(92 + " + (Math.abs(m_FrameOffset)) + ") & -8");
		
		
	}

	//this does the normal function call in asmbly, set all params and call
	//pass null for param 'this' of current struct if no struct
	public void
	AsmblyCall(Vector p, STO ret){
		if (m_isStructMember)
			AsmblyCall(p, ret, m_memberOfStruct);
		else
			AsmblyCall(p, ret, null);
	}

	//this does the function call in asmbly, set all params and call
	public void
	AsmblyCall(Vector p, STO ret, STO struct_this){
		
		//call the func in FuncProtoType, giving it our label to call
		getProtoType().AsmblyCall(p, ret, struct_this, m_label);
		

	}
	
	//this does the function return in asmbly, set return val and ret restore
	public STO
	AsmblyReturn (STO ret){

		//make sure ret is loaded, and load into I0
		if (!getReturnType().isVoid() && ret != null){
			if (getReturnType().isFloat()){
				//see if we need to convert
				if (ret.getParentType().isInt()){
					//promote src to float by using the cast op
					ret = MyParser.DoBinaryExpr(new EmptySTO(new FloatType()), new CastOp(), ret);
				}
				//and load up
				ret.AsmblyLoad(F0);
			}else
				ret.AsmblyLoad(I0);
		}
		//and ret and return
		//write ret and restore
		AssemblyWriter.writeSmtAndComment("Jump back", RET);
		AssemblyWriter.writeSmtAndComment("and restore SP", RESTORE);
		//and clear
		ret.AsmblyClearReg();
		
		return ret;
	}

	// ----------------------------------------------------------
	// this function saves the lowest SP in a global var called  'GLOB.LowSP'
	public static void
	saveLowestSP(){
		String overwriteLabel = AssemblyWriter.getNextCmpTempLabel() + ".lower";
		String endLabel = AssemblyWriter.getNextCmpTempLabel() + ".end";
		
		
		//set label
		AssemblyWriter.writeSmtAndComment ("set label of global lowest SP val into G1", SET, "GLOB.LowSP", G1);
		//and load into G2
		AssemblyWriter.writeSmtAndComment ("load", LOAD, G1, G2);
		
		//compare
		AssemblyWriter.writeSmtAndComment ("compare", CMP, SP, G2);
		
		
		//and branch
		AssemblyWriter.writeSmtAndComment ("branch if higher", BGU, endLabel);
		
		//and store
		AssemblyWriter.writeSmtAndComment ("store current SP to global var", STORE, SP, G1);
		
		//end
		AssemblyWriter.writeSmt (LABEL, endLabel);
		
		
	}

}
