package cs131w.symtab.sto;
//---------------------------------------------------------------------
//
//---------------------------------------------------------------------
import  cs131w.symtab.types.*;
import  cs131w.assembly.*;
import  cs131w.parser.MyParser;

public class ParamSTO extends VarSTO
{
	private boolean m_isRef;
	
	//----------------------------------------------------------------
	//
	//----------------------------------------------------------------

	public 
	ParamSTO ()
	{
		this ("Param", null);
	}

	public 
	ParamSTO (String strName, Type t)
	{
		this (strName, t, false);
	}

	public 
	ParamSTO (String strName, Type t, boolean ref)
	{
		super (strName, t);
		m_isRef = ref;
		asmbl_created = true;
		m_label = "parm." + m_strName + "$L" + MyParser.GetLineNum();
		
	}


	public boolean
	getIsRef(){
		return m_isRef;
	}

	public boolean
	isRef(){
		return m_isRef;
	}
	
	//----------------------------------------------------------------
	//
	//----------------------------------------------------------------
	public boolean   
	isParam () 
	{
		return true;
	}
	
	
	public String
	toString()
	{
		String optref = ((m_isRef) ? "&" : "");
		if (MyParser.DEV){
			String ad = "";
			if (isInReg())
				ad = m_register;
			else
				ad = getAddrStr();
				
			return ("(Param)" + m_type + " " + optref + m_strName + "@" + ad);
		}else
			return (m_type + " " + optref + m_strName);
		
	}
	
	//gets the mangled name for function overloading, just return type info in this case
	public String
	getMangledName()
	{
		return m_type.getRootType().getMangledName();
	}
	
	public Object 
	clone(){		// throws CloneNotSupportedException {
		
		ParamSTO ret = new ParamSTO(new String(m_strName), (Type)m_type.clone(), m_isRef);
		ret.setIsModifiable(m_isModifiable);
		ret.setIsAddressable(m_isAddressable);
		//asmbly
		ret.setReg(m_register);
		ret.setMemBase(mem_base);
		ret.setMemOffset(mem_offset);
		ret.setIsLocal(mem_isLocal);
		if (m_static) ret.setStatic();
		if (asmbl_created) ret.setAsmblyCreated();
		if (hasRuntimeAddr()) ret.setRuntimeAddr((StorageLocation)mem_runtimeAddr.clone());
		
		return ret;
	}
		
	public String getName(){
		return m_strName;
	}
	
	
	//we just need to overwrite funcs for storing an loading this param
	//since it might be a ref
	public void
	AsmblyLoad(){ 
		if (isError()) return;
		if (isInReg()) return;
		
		String in;
		if (getParentType().isFloat())
			in = Regs.getNextFloatReg();
		else
		 	in = Regs.getNextReg();
		
		this.AsmblyLoad(in);
	}
	
	//loads this STO into a reg, if its in mem, do care about where it is stored into
	public void
	AsmblyLoad(String into){ 
		MyParser.debug(this + ".AsmblyLoad(into) called in Param.");
		if (isError()) return;
		
		//now we just check if its a ref, if so we load the addr into L7, using parents method
		//and then load from that address
		if (m_isRef){
			//load the address from param into L7
			super.AsmblyLoad(L7);
			//and load
			AssemblyWriter.writeSmt(LOAD, L7, into);
			m_register = into;
		}else{	//not a ref, could be array or struct as well
			//not a ref, just load normally from where its stored on the stack (%fp+...)
			super.AsmblyLoad(into);
		}
		
	}
	
	//load an address into specified reg, might be ref, so we need to overload
	public void
	AsmblyLoadAddr(String into){
		MyParser.debug(this + ".AsmblyLoadAddr("+into+") called.");
		if (isError()) return;
		if (!asmbl_created) AsmblyCreate();
	
		if (hasRuntimeAddr()){
			loadRuntimeAddrInto(into);
		}else{
			//now if its a ref, array or struct, we just load the addr
			if (m_isRef || getType().isStruct() || getType().isArray()){
				//just load the ref into the outreg
				super.AsmblyLoad(into);
			}else{
				//if not load the addr
				super.AsmblyLoadAddr(into);
			}
		}
	} 
	
	
	
	//stores this STO back from reg into mem where it was created, and thus frees this reg
	public void
	AsmblyStore(){ 
		MyParser.debug(this + ".AsmblyStore called in Param.");

		if (isError()) return;
		//if not loaded: return
		if (!isInReg()) return;
		
		//store away
		if (m_isRef){
			//load the address from param into L7
			//super.AsmblyLoad(L7);
			//do not use super's method, since its just going to move it, from where it is loaded
			AssemblyWriter.writeSmtAndComment ("Load Addr of '" + this + "' into L7", LOAD, getAddrStr(), L7);
			//and store
			AssemblyWriter.writeSmtAndComment ("and store away", STORE, m_register, L7);
		}else{
			//not a ref, just store normally to where its stored on the stack (%fp+...)
			super.AsmblyStore();
		}
					
		Regs.FreeReg(m_register);
		m_register = null;
	
	}
	
	
}
