package cs131w.symtab.sto;
//---------------------------------------------------------------------
//
//---------------------------------------------------------------------
import  cs131w.symtab.types.*;
import  cs131w.error.*;
import  cs131w.assembly.*;
import  cs131w.symtab.*;
import  cs131w.util.*;
import  cs131w.parser.MyParser;
import 	java.util.Vector;

public class UeberFuncSTO extends FuncSTO
{
		protected Scope				m_OverloadScope;
		protected Type				m_retType;
		protected static FuncSTO	m_currentFunc;
	//----------------------------------------------------------------
	//
	//----------------------------------------------------------------
	public 
	UeberFuncSTO ()
	{
		this("Ueber");
	}

	public 
	UeberFuncSTO (String strName)
	{
		this(strName, new VoidType());
	}
	
	public 
	UeberFuncSTO (String strName, Type t){
		this(strName, t, 0, null);
	}
	
	public
	UeberFuncSTO(String strName, Type t, int deep, STO func)
	{
		
		//just set the type to the void type and a proto to scope
		super (strName, new VoidType());
		MyParser.debug("ctoring UeberFuncSTO w/ "+strName+",type:"+t+", and func "+func);

		setRVal();
		m_scopeDeepness = deep;
		if (t.isFunction())
			m_retType = ((FuncProtoType)t).getReturnType();
		else 
			m_retType = t;
			
		m_OverloadScope = new Scope();
		
		if (func != null)
			addOverload(func);
	}



	//----------------------------------------------------------------
	//
	//----------------------------------------------------------------
	public boolean
	isFunc () 
	{ 
		return true;
   	}

	public boolean
	isUeberFunc () 
	{ 
		return true;
   	}


	//----------------------------------------------------------------
	//
	//----------------------------------------------------------------
	public boolean
	isOverloaded () 
	{ 
		return (m_OverloadScope.getCount() > 1);
   	}


	//----------------------------------------------------------------
	//
	//----------------------------------------------------------------
	public STO
	accessLocal(String id){
		return m_OverloadScope.accessLocal(id);
	}

	public STO
	access(String id){
		return m_OverloadScope.accessLocal(id);
	}

	public STO
	getSingle(){
		return m_OverloadScope.getSingle();
	}
	
	//----------------------------------------------------------------
	//
	//----------------------------------------------------------------
	public void
	addOverload (STO sto)
	{
		m_OverloadScope.InsertLocal(sto);
	}
	
	
	//----------------------------------------------------------------
	//
	//----------------------------------------------------------------
	public Type
	getReturnType ()
	{
		return m_retType;
	}

	public void
	setReturnType (Type t)
	{
		m_retType = t;
	}

	public FuncProtoType
	getProtoType ()
	{
		if (!isOverloaded())
			return ( ((FuncSTO)getSingle()).getProtoType() );
		else
			return null;
	}

	public Type
	getType ()
	{
		return getProtoType ();
	}



	//----------------------------------------------------------------
	//
	//----------------------------------------------------------------
	public Scope
	getScope ()
	{
		return m_OverloadScope;
	}

	public void
	setScope (Scope s)
	{
		m_OverloadScope = s;
	}


	//----------------------------------------------------------------
	// If we have found a return already 
	//----------------------------------------------------------------
	public void
	ReturnFound ()
	{
		m_bReturnFound = true;
	}

	public boolean
	getReturnFound ()
	{
		return m_bReturnFound;
	}

	public void
	setReturnFound (boolean f)
	{
		m_bReturnFound = f;
	}


	//----------------------------------------------------------------
	//----------------------------------------------------------------
	public String
	toString()
	{
		if (MyParser.DEV)
			return ("(UFunc)" + m_strName + "():" + m_type);
		else
			return (m_strName);
		
	}	

	public Object 
	clone(){
		//UeberFuncSTO(String strName, Type t, int deep, STO func)
		
		UeberFuncSTO ret = new UeberFuncSTO(new String(m_strName), (Type)m_type.clone(), m_scopeDeepness, null);
		ret.setScope((Scope) m_OverloadScope.clone());
		ret.setReturnType(m_retType);
		ret.setIsModifiable(m_isModifiable);
		ret.setIsAddressable(m_isAddressable);
		//now set the asmbly vars, too
		ret.setReg(m_register);
		ret.setMemBase(mem_base);
		ret.setMemOffset(mem_offset);
		ret.setIsLocal(mem_isLocal);
		if (m_static) ret.setStatic();
		if (asmbl_created) ret.setAsmblyCreated();
		if (hasRuntimeAddr()) ret.setRuntimeAddr((StorageLocation)mem_runtimeAddr.clone());
		
		return ret;
	}

	//----------------------------------------------------------------
	//----------------------------------------------------------------
	public STO
	DoFuncCall (Vector p){
		
		//first we need to build the actual mangled function name from the supplied params
		String mangledName = Utils.getMangledNameFromVector(p);
		
		MyParser.debug ("overloaded-funcall calling: " + this + ", mangled: " + mangledName + " with "+p);
		
		//now check if that name is there
		STO funcsto; 
		
		if (isOverloaded()){
			funcsto = access(mangledName);
			if (funcsto == null)
				return new ErrorSTO (this, Formatter.toString(ErrorMsg.error22_Illegal, this.getName()));
		
		}else{	//its not overloaded, so just get single func that's in this ueberFuncSTO
			funcsto = getSingle();
		}	
		
		//and reset the current struct just in case
		if (m_isStructMember){
			funcsto.setIsStructMember(m_memberOfStruct);
		}
		
		return funcsto;
		
	}


	//----------------------------------------------------------------
	//----------------------------------------------------------------
	public static
	STO start_create(String id, Type t){
		
		SymbolTable st = MyParser.getSymTab();
		MyParser.SaveLineNum ();
		STO ret;
		
		String mangled = t.getMangledName();
		
		//create the func early
		STO sto = new FuncSTO (mangled, t, st.getLevel()+1, id);
		
		//first check to see if its already there with another name
		STO old = st.accessLocal(id);
		m_currentFunc = (FuncSTO) sto;
		st.setFunc (sto);
		
		//if it is and not a func, throw a redecl error
		if (old != null && !old.isUeberFunc() ){
			return MyParser.STO_getRedeclIDReport (id);
		}
		
		UeberFuncSTO oldfunc = (UeberFuncSTO) old;

		//first check if its already there, and overloaded
		if (oldfunc != null){
			//now check if that particular func already exists in there
			if (oldfunc.access(mangled) != null){
				return new ErrorSTO (sto, Formatter.toString(ErrorMsg.error22_Decl, id));
			}else{
				//now we just add it here, to the ueberfuncsto
				oldfunc.addOverload(sto);
				//and return an empty sto to not add it to the ST any more
				ret = new EmptySTO();
			}
			
		}else{
			//not there yet, create new
			//UeberFuncSTO(String strName, Type t, int deep, STO func)
			ret = new UeberFuncSTO(id, t, st.getLevel()+1, sto);
			//set struct into it, if defining
			
			if (st.isDefiningStruct()){
				//just remember for later, that we are a member of a struct
				ret.setIsStructMember(st.getStruct());
				sto.setIsStructMember(st.getStruct());
				
			}

			
			oldfunc = (UeberFuncSTO)ret;
		}
		
		//and set our struct mem flag into it, if we are mem of struct, to pass this* in o0
		if (oldfunc.m_isStructMember){
			sto.setIsStructMember(oldfunc.m_memberOfStruct);
		}
		
		//now start the asmbly output
		sto.AsmblyCreate();
		
		return ret;
	}


	public static
	STO finish_parsing(){

		STO func = m_currentFunc;
		
		MyParser.debug ("DoFuncDelc_2 doing end of body for "+func);
		
		
		if (func == null)
		{
				MyParser.debug ("internal: DoFuncDecl_2 says no func!");
				return new EmptySTO();
		}
		
		if (func.isError())
			func = ((ErrorSTO)func).getBase();
		
		FuncSTO f = (FuncSTO)func;
		
		if (!f.getReturnFound() && !f.getReturnType().isVoid()){
			MyParser.debug("DoFuncDelc_2 checking for return in: "+func+". return found is: "+f.getReturnFound() + ". RetType.void: "+f.getReturnType().isVoid());
			//STO_doErrorReport (f, ErrorMsg.error6b_Return_missing, true);
			return new ErrorSTO (f, ErrorMsg.error6b_Return_missing);
		}
		
		m_currentFunc = null;
		
		f.AsmblyFinish();
		return func;

	}

}
