package cs131w.symtab.sto;
//---------------------------------------------------------------------
//
//---------------------------------------------------------------------
import  cs131w.symtab.types.*;
import  cs131w.parser.MyParser;
import  cs131w.error.*;
import  cs131w.assembly.*;
import  cs131w.symtab.*;

public class VarSTO extends AdressableSTO
{
	
	private String stat_init_flag = "ERROR";
	private String skip_stat_init_label = "ERROR";
	
	
	
	//----------------------------------------------------------------
	//
	//----------------------------------------------------------------
	public 
	VarSTO ()
	{
		this ("var");
		
	}

	public 
	VarSTO (String strName)
	{
		this (strName, null);
		
	}

	public 
	VarSTO (String strName, Type t)
	{
		super (strName, t);
		setModLVal();
		
		//check if its an array, coz that shouldnt be modifiable
		if (t != null && t.isArray())
			setIsModifiable(false);
			
		m_label = "var." + m_strName + "$L" + MyParser.GetLineNum();
		
	}

	//----------------------------------------------------------------
	//
	//----------------------------------------------------------------


	public boolean   
	isVar () 
	{
		return true;
	}
	
	public String
	toString()
	{
		if (MyParser.DEV){
			String ad = "";
			if (isInReg())
				ad = m_register;
			else
				ad = getAddrStr();
				
			return ("(Var)" + m_strName + ":" + m_type + "@" + ad);
		}else
			return (m_strName);
		
	}	

	//----------------------------------------------------------------
	// Factory method to do the checking and make a new VarSTO for us
	//----------------------------------------------------------------

	public static STO
	create (String id, Type t){
		
			//just make the new Varsto
			VarSTO var = new VarSTO (id, t);
			SymbolTable st = MyParser.getSymTab();
			
			if (st.accessLocal (id) != null)
			{
				//STO_doRedeclIDReport (id);
				return MyParser.STO_getRedeclIDReport(id);
			}

			//check if we are defining the struct, we are trying to use as a type here
			if (st.isDefiningStruct()){
				STO struct = st.getStruct();
				TypedefSTO tsto;
				
				if (struct.isError())
					tsto = (TypedefSTO) (((ErrorSTO)struct).getBase());
				else
					tsto = (TypedefSTO)struct;
				
				if (t == tsto.getParentType()){
					return new ErrorSTO (tsto, Formatter.toString(ErrorMsg.error13b_Struct, id));
				}
				
			} 
		
		return var;
		
	}
	
	
	//----------------------------------------------------------------
	//	Assembly functions
	//----------------------------------------------------------------

	//this creates and sends to the writer the assembly code for creating and storing this const
	public void AsmblyCreate(STO sto){ 
		if (asmbl_created) return;
		
		if (sto != null && sto.isConst())
			AsmblyCreate((ConstSTO)sto); 
		else
			AsmblyCreate(null); 
	} 

	public void AsmblyCreate(){ AsmblyCreate(null); } 

	public void
	AsmblyCreate(ConstSTO init){ 
		if (asmbl_created) return;
		asmbl_created = true;
	
		if (getParentType().isArray() || getParentType().isStruct()){
			if (MyParser.isInGlobalSpace() || m_static){
				AsmblyCreateSpaceGlobal();
			}else{
				this.AsmblyCreateTemp(null);
			}
		}else{
		
			if (MyParser.isInGlobalSpace() || m_static){
				AsmblyCreateVarGlobal(init);
			}else{
				this.AsmblyCreateTemp(init);
			}
		}
		
	}

	public void
	AsmblyCreateVarGlobal(){ this.AsmblyCreateVarGlobal(null); }	
	
	//this functions emits code for making this global var
	public void
	AsmblyCreateVarGlobal(ConstSTO init){
//		MyParser.debug(this + ".emitGlobalAssemblyCalled.");
		markGlobal();
		
		//switch to rodata section
		MyParser.debug(this + ".AsmblyCreateVarGlobal switching to data");
		AssemblyWriter.switchSection(DATA);
		
		if (!m_static){
			//write global marker
			MyParser.debug(this + ".AsmblyCreateVarGlobal writing global marker");
			AssemblyWriter.writeSmt(GLOBAL, getLabel());
		}
		
		
		//write the var
		MyParser.debug(this + ".AsmblyCreateVarGlobal writing var");
		String val = ( (init != null) ? init.getValueAsString() : "0");
		String var_init = (getParentType().isFloat()) ? FLOAT_VAR_INIT : WORD_VAR_INIT;
		
		
		AssemblyWriter.writeSmt(var_init, getLabel(), val);
		AssemblyWriter.switchSection(TEXT);
		
		//fields are already set by markGlobal
	}
	
	//this just creates some space in the BSS if this is a struct or array
	public void AsmblyCreateSpaceGlobal(){
		markGlobal();
		
		if (MyParser.isInGlobalSpace()){
			AssemblyWriter.writeSmt(GLOBAL, getLabel());
		}
		
		//switch to section BSS to create some 0 filled space
		AssemblyWriter.switchSection(BSS);
		
		AssemblyWriter.writeSmt(LABEL, getLabel());
		String s = new Integer(getParentType().getSize()).toString();
		AssemblyWriter.writeSmtAndComment ("create space for "+s+" bytes", SKIP, s);
		AssemblyWriter.switchSection(TEXT);
		
	}



	//this creates and stores a temp value in mem somewhere
	public void
	AsmblyCreateTemp(ConstSTO init){ 
		
		int off = MyParser.increaseFrameOffset(getSize());
		markLocal(off);
		
		//output some comment to denote the offset of this var in the asmbly
		MyParser.asmblyWriteVarSmt (COMMENT, "creating local var '"+this+"' at offset %fp" + off);
		
		//not initialized just return
		if (init == null) return;
		
		//this means its a const init

		
		//if this is float we need to store the float val in a label
		if (getParentType().isFloat()){
			
			MyParser.debug(this + ".AsmblyCreateTemp writing float var");
			
			String lbl = AssemblyWriter.getNextTempLabel();
			
			AssemblyWriter.switchSection(DATA);
			//AssemblyWriter.writeSmt(LABEL, lbl);
			AssemblyWriter.writeSmt( FLOAT_VAR_INIT, lbl, init.getValueAsString() );
			
			AssemblyWriter.switchSection(TEXT);
			
			//now load it back up from its label
			MyParser.asmblyWriteVarSmt (SET, lbl, L7);
			MyParser.asmblyWriteVarSmt (LOAD, L7, F0);
			setReg(F0);
		
		}else{
			MyParser.debug(this + ".AsmblyCreateTemp writing bool or int var");
			//else this means its a bool or int
			//we just set it somewhere and store it in temp
			//make sure to call the MyParser func here, which either sets it into current place or global.INIT if we are in global space
			MyParser.asmblyWriteVarSmt (SET, init.getValueAsString(), L6);
			setReg(L6);
		}
		
		MyParser.debug(this + ".AsmblyCreateTemp storing var");
		//now store the puppy into its actual location
		AsmblyStore();
		
		
	}

	//----------------------------------------------------------------
	//  funcs for doing the guard against doing static inits more than once
	//----------------------------------------------------------------
	public void
	AsmblyStartStaticInitGuard(){
		if (!m_static)
			return;
		
			stat_init_flag = getLabel() + ".static_flag";
			skip_stat_init_label = "skip_" + stat_init_flag;
			
			//create flag
			AssemblyWriter.switchSection(DATA);
			AssemblyWriter.writeSmt(WORD_VAR_INIT, stat_init_flag, "0");
			AssemblyWriter.switchSection(TEXT);
			
			//now check
			AssemblyWriter.writeSmt(SET, stat_init_flag, L0);
			AssemblyWriter.writeSmt(LOAD, L0, L1);
			AssemblyWriter.writeSmt(CMP, L1, G0);
			
			//and branch
			AssemblyWriter.writeSmt(BNE, skip_stat_init_label);
			AssemblyWriter.writeSmt(NOP);
			
	}
	
	public void
	AsmblyFinishStaticInitGuard(){
	
		if (!m_static)
			return;
		
		//and do the end of stat init code

			//set flag to 1
			AssemblyWriter.writeSmt(SET, "1", L0);
			AssemblyWriter.writeSmt(SET, stat_init_flag, L1);
			AssemblyWriter.writeSmt(STORE, L0, L1);
			//and write label
			AssemblyWriter.writeSmt(LABEL, skip_stat_init_label);
		
	
	}
		
	//----------------------------------------------------------------
	//
	//----------------------------------------------------------------

	public void
	AsmblyCIn(){ 
		MyParser.debug(this + ".AsmblyCIn called.");

		if (getParentType().isInt()){
			
			//how do we call this?
			//inputInt();
			
			AssemblyWriter.writeSmt(CALL, "inputInt");
			AssemblyWriter.writeSmt(NOP);
			//AssemblyWriter.writeSmt(MOVE, O0, L0);
			//and signal that the return value is in l0
			setReg(O0);
			
		}else if (getParentType().isFloat()){
			
			AssemblyWriter.writeSmt(CALL, "inputFloat");
			AssemblyWriter.writeSmt(NOP);
			//and signal that the return value is in l0
			setReg(F0);
			
		}
		AsmblyStore();
		

	}
	
}
