package cs131w.symtab.types;
import cs131w.parser.MyParser;


public class AliasType extends Type
{
	
	//the base/parent type
	protected Type		m_baseType;
	protected Type		m_rootType;
	
	//----------------------------------------------------------------
	//
	//----------------------------------------------------------------
	public 
	AliasType ()
	{
		this(null);
	}
	
	public 
	AliasType (Type base){
		this("alias", base);
	}
	
	public 
	AliasType (String strName, Type base)
	{
		super(strName,0);
		m_baseType = base;
		if (!m_baseType.isAlias())
			m_rootType = m_baseType;
		else
		 	m_rootType = ((AliasType)m_baseType).getRootType();

	}


	//ptrs actually have state, so override clone
	public Object 
	clone(){
		AliasType ret = new AliasType(new String(m_typeName), (Type)m_baseType.clone());
		return ret;
	}
	
	//----------------------------------------------------------------
	// define own getClass method, since getClass in Object is final
	// and thus can't be overriden in AliasType
	//----------------------------------------------------------------
	public Class
	getClazz(){
		return getRootType().getClazz();
	}

	//----------------------------------------------------------------
	// methods to get base stuff
	//----------------------------------------------------------------
	public Type
	getRootType(){
		return m_rootType;
	}

	public int
	getBaseSize(){
		return m_rootType.getSize();
	}

	//----------------------------------------------------------------
	// and redefine funcs to call base things, size
	//----------------------------------------------------------------
	public int
	getSize ()
	{
		return getBaseSize();
	}

	public boolean isEquivalent(Type t){
		if (t == null) return false;
		return (m_rootType.getClazz() == t.getClazz());
	}

	public boolean isAssignable(Type t){
		return m_rootType.isAssignable(t);
	}	


	//----------------------------------------------------------------
	//----------------------------------------------------------------
	public boolean	isAlias ()	{ return true; }


	//----------------------------------------------------------------
	//
	//----------------------------------------------------------------
	public String
	toString()
	{
		if (MyParser.DEV)
			return (m_typeName + "(Alias:" + m_baseType + ")");
		else
			return m_typeName;
	}

	//gets the mangled name for function overloading, just return type info in this case
	public String
	getMangledName()
	{
		return m_rootType.getMangledName();
	}

	//----------------------------------------------------------------
	//  need to overwrite all is*() Methods to reflect baseType
	//----------------------------------------------------------------

	//first level hierarchy
	public boolean	isBasic ()		{ return m_rootType.isBasic(); }
	public boolean	isVoid ()		{ return m_rootType.isVoid(); }
	public boolean	isComposite ()	{ return m_rootType.isComposite(); }
	public boolean	isError ()		{ return m_rootType.isError(); }

//	public boolean	isNull ()		{ return m_rootType.isNull(); }
	public boolean	isFunction ()	{ return m_rootType.isFunction(); }
	
	//second level hierarchy
	//basic
	public boolean	isNumeric ()	{ return m_rootType.isNumeric(); }
	public boolean	isBool ()		{ return m_rootType.isBool(); }
	public boolean	isChar ()		{ return m_rootType.isChar(); }

	//comp
	public boolean	isArray ()		{ return m_rootType.isArray(); }
	public boolean	isStruct ()		{ return m_rootType.isStruct(); }
	public boolean	isPtr ()		{ return m_rootType.isPtr(); }
	public boolean	isString ()		{ return m_rootType.isString(); }

	//third level hierarchy
	//numeric
	public boolean	isInt ()		{ return m_rootType.isInt(); }
	public boolean	isFloat ()		{ return m_rootType.isFloat(); }

	//pointer
	public boolean	isBasicPtr ()	{ return m_rootType.isBasicPtr(); }
	public boolean	isFuncPtr ()	{ return m_rootType.isFuncPtr(); }

	//get the number of derefs to the rootType if pointer
	public int		getRootDerefs(){
		return m_baseType.getRootDerefs();
	}
	
	public boolean	isCastable (Type t)	{ return m_rootType.isCastable(t); }



	
	
}