package cs131w.symtab.types;

import cs131w.parser.MyParser;
import cs131w.symtab.sto.*;
import cs131w.error.*;

public class ArrayType extends CompType
{
	//dimension
	private int			m_dim;
	//base type
	private Type		t_baseType;
	
	//----------------------------------------------------------------
	//
	//----------------------------------------------------------------
	public 
	ArrayType ()
	{
		this(0);
	}

	public 
	ArrayType (int dim)
	{
		this(new VoidType(),dim);
	}
	
	public 
	ArrayType (Type baseType, int dim)
	{
		super("array");
		setUnderlyingType(baseType);
		setDim(dim);
		setSize(dim * baseType.getSize());
	}


	//arrays actually have state, so override clone
	public Object 
	clone(){
		ArrayType ret = new ArrayType((Type)t_baseType.clone(), m_dim);
		return ret;
	}
	
	// -----------------------------------------
	// Dimension
	// -----------------------------------------
	public int
	getDim ()
	{
		return m_dim;
	}

	private void
	setDim (int dim)
	{
		m_dim = dim;
		setSize(m_dim * t_baseType.getSize());
	}

	// -----------------------------------------
	// BaseType
	// -----------------------------------------
	public Type
	getUnderlyingType ()
	{
		return this.t_baseType;
	}

	public void
	setUnderlyingType (Type t)
	{
		t_baseType = t;
		setSize(m_dim * t_baseType.getSize());
	}
	
	//----------------------------------------------------------------
	//
	//----------------------------------------------------------------
	public String
	toString()
	{
		return t_baseType + "[" + m_dim + "]";
	}
	
	public String
	getMangledName()
	{
		return this.toString();
	}
	
	//----------------------------------------------------------------
	//----------------------------------------------------------------
	public boolean isEquivalent(Type t){
		if (t == null || !t.isArray()) return false;
		
		ArrayType other = (ArrayType) t.getRootType();
		
		
		if (other.getUnderlyingType().getClazz() == t_baseType.getClazz() && other.getDim() == m_dim){
			return true;
		}
		
		return false;
	}
	
	//we can assign an array to a basic pointer of same type, so check that here 
	public boolean isAssignable(Type t){
		
		if (	t.isPtr() 
			&& 
				( ((PtrType)t.getRootType()).getPointedToType().getClazz() == t_baseType.getClazz() )
			&& 
				(((PtrType)t.getRootType()).getDerefs() == 1)
			)
			return true;
			
		return isEquivalent(t);
	}
	

	public static
	Type create(STO sto){
		
		if (sto.isError() || !sto.isConst()){
			
			return new ErrorType (ErrorMsg.error10c_Array);
		
		//check if its an int or not const, has to be
		}else if (!sto.getType().isInt()){
		
			return new ErrorType (Formatter.toString(ErrorMsg.error10i_Array, sto.getType()));
		
		}else if ( ((ConstSTO)sto).getValue() <= 0){
	
			return new ErrorType (Formatter.toString(ErrorMsg.error10z_Array, ((ConstSTO)sto).getIntValue()));
	
		}else
			return (new ArrayType(((ConstSTO)sto).getIntValue()));  
		
	}


	//----------------------------------------------------------------
	//----------------------------------------------------------------
	public boolean	isArray ()		{ return true; }
	
	
}