package cs131w.symtab.types;
import cs131w.parser.MyParser;
import cs131w.util.*;


public abstract class PtrType extends CompType
{
	
	//----------------------------------------------------------------
	//	number of *'s in the definition, has to be >= 1 since its a ptr
	//----------------------------------------------------------------
	protected int			m_numDeref = 0;
	protected Type			m_baseType;




	public 
	PtrType (String strName, int size)
	{
		this(strName, size, null, 0);
	}

	public 
	PtrType (String strName, int size, Type baseT, int derefs)
	{
		super(strName,size);
		m_numDeref = derefs;
		m_baseType = baseT;
	}


	//----------------------------------------------------------------
	//----------------------------------------------------------------
	public String
	toString()
	{
		return ("(Ptr)" + m_baseType + Utils.strRepeat("*",m_numDeref));
	}

	public String
	getMangledName()
	{
		return (m_baseType.getRootType().getMangledName() + Utils.strRepeat("*",m_numDeref));
	}
	

	
	//----------------------------------------------------------------
	//----------------------------------------------------------------
	public Type
	getPointedToType ()
	{
		return	m_baseType;
	}

	protected void
	setPointedToType (Type type)
	{
		m_baseType = type;
	}

	//----------------------------------------------------------------
	// define own getClass method, since getClass in Object is final
	// and thus can't be overriden in AliasType
	//----------------------------------------------------------------
	public Class
	getClazz(){
		return m_baseType.getClazz();
	}

	//----------------------------------------------------------------
	//----------------------------------------------------------------
	public int
	getDerefs ()
	{
		return	m_numDeref;
	}

	public void
	setDerefs (int d)
	{
		m_numDeref = d;
	}
	
	public int getRootDerefs () { 
		return (getDerefs() + m_baseType.getRootDerefs()); 
	}

	//----------------------------------------------------------------
	//----------------------------------------------------------------
	public boolean isEquivalent(Type t){
		
		MyParser.debug(this + ".isEquiv() to " + t);
		
		if (t == null || !t.isPtr()) return false;
		
		PtrType other = (PtrType) t.getRootType();
		
		MyParser.debug(this + ".checking Clazz: this.clz=" + m_baseType.getClazz() + ", o.clz=" + other.getPointedToType().getClazz());
		MyParser.debug(this + ".checking derefs: this.dr=" + this.getRootDerefs() + ", o.dr=" + other.getRootDerefs());
		if (	(other.getPointedToType().getClazz() == m_baseType.getClazz()) 
			&& 
				//(other.getDerefs() == m_numDeref)
				(other.getRootDerefs() == this.getRootDerefs())
		   )
			return true;

		MyParser.debug(this + ".eqv: checking Array assignability ");
		//array name is a pointer to first element, so check that here
		if (	other.getPointedToType().isArray()
			&& 
				(other.getPointedToType().getUnderlyingType().getClazz() == m_baseType.getClazz() ) 
			&&
				(other.getRootDerefs() == (this.getRootDerefs() - 1))
		   )
			return true;

		MyParser.debug(this + ".eqv: checking Array assignability switched");
		if (	this.getPointedToType().isArray()
			&& 
				(other.getPointedToType().getClazz() == m_baseType.getUnderlyingType().getClazz() ) 
			&&
				((other.getRootDerefs() - 1) == this.getRootDerefs() )
		   )
			return true;
		
		MyParser.debug(this + ".eqv: returning false!");
			
			
		return false;
	}

	public boolean isAssignable(Type t){
		MyParser.debug(this + ".isAssignable() to " + t);
		
		if (t == null || !t.isPtr()) return false;
		PtrType other = (PtrType) t.getRootType();
		
		MyParser.debug(this + ".isAssignable() to Void");
		//any pointer type is assignable to a void*
		if (other.getPointedToType().isVoid() || this.getPointedToType().isVoid())
			return true;
		
		MyParser.debug(this + ".isAssignable() to: checking struct!");
		//if this is a struct, it has to use loose name equiv. so check that
		if (other.getPointedToType().isStruct() && this.getPointedToType().isStruct()){
			StructType st = (StructType) this.getPointedToType().getRootType();
			StructType so = (StructType) other.getPointedToType().getRootType();
			MyParser.debug(this + ".isAssignable() checking both structs, this:" + st + ", o:" + so);
			
			return st.isAssignable(so);
			
		}
		
		boolean equiv = isEquivalent(t);
		
		MyParser.debug(this + ".isAssignable() to: returning isEquiv=" + equiv);
		
		return equiv;
	}	



	//----------------------------------------------------------------
	//----------------------------------------------------------------
	public boolean	isPtr ()		{ return true; }
	
	
}