package cs131w.symtab.types;

import java.util.Vector;
import cs131w.symtab.*;
import cs131w.symtab.sto.*;
import cs131w.parser.MyParser;


public class StructType extends CompType
{

	//Scope of variables
	private Scope		 		m_varScope;

	//the tag of the parent typedef (for overloading purposes repeated here)
	private String		 		m_TagName;
	
	//----------------------------------------------------------------
	//
	//----------------------------------------------------------------
	public 
	StructType ()
	{
		this("tag");
	}

	public 
	StructType (String tag)
	{
		this(new Scope(), tag);
	}

	public 
	StructType (Scope s, String tag)
	{
		super("struct",0);
		m_varScope = s;
		m_TagName = tag;
	}


	//ptrs actually have state, so override clone
	public Object 
	clone(){
		StructType ret = new StructType((Scope) m_varScope.clone(), new String(m_TagName));
		return ret;
	}

	// -----------------------------------------
	// Count of members
	// -----------------------------------------
	public int
	getCount ()
	{
		return m_varScope.getCount();
	}

	public int
	getSize ()
	{
		return m_varScope.getSize();
	}

	// -----------------------------------------
	// tag
	// -----------------------------------------
	public String
	getTag ()
	{
		return m_TagName;
	}

	public void
	setTag (String t)
	{
		m_TagName = t;
	}

	// -----------------------------------------
	// Scope
	// -----------------------------------------
	public Scope
	getScope ()
	{
		return m_varScope;
	}

	public void
	setScope (Scope s)
	{
		m_varScope = s;
	}

	//this adds a member to the scope and struct
	public void
	addMember (STO sto)
	{
		//insert into scope
		m_varScope.InsertLocal (sto);
		//and update size
		m_size += sto.getSize();
		
	}
	
	public STO
	access (String strName)
	{
		return	m_varScope.access (strName);
	}

	
	public boolean isEquivalent(Type t){
		if (t == null || !t.isStruct() || this.getClazz() != t.getClazz()) return false;
		boolean ret = true;
		StructType s = (StructType) t.getRootType();
		if (!m_varScope.equals(s.getScope()))
			ret = false;
		
		MyParser.debug(this + ".checking equiv to: " + t + ". returning: "+ret);
		 
		return ret;
	}

	//should only be assignable if exact same type
	public boolean isAssignable(Type t){
		MyParser.debug(this + ".checking assign to: " + t);
		if (t == null || !t.isStruct() || this.getClazz() != t.getRootType().getClazz()) return false;
		return true;
	}	

	
	//----------------------------------------------------------------
	//----------------------------------------------------------------
	public String
	toString(){
		if (!MyParser.DEV)
			//return m_typeName;
			return m_TagName;
		
		return ("struct("+m_varScope.toSpecialString(this)+")");
	}	

	public String
	getMangledName()
	{
		return ("struct$" + m_TagName);
	}


	//----------------------------------------------------------------
	//----------------------------------------------------------------
	public boolean	isStruct ()		{ return true; }
	
	
}
