package cs131w.symtab.types;

//---------------------------------------------------------------------
// This is the top of the Type hierarchy. You most likely will need to
// create sub-classes (since this one is abstract) that handle specific
// types, such as IntType, FloatType, ArrayType, etc.
//---------------------------------------------------------------------


public abstract class Type implements Cloneable
{
	//----------------------------------------------------------------
	//	Name of the Type (e.g., int, bool, or some typedef
	//----------------------------------------------------------------
	protected String  		m_typeName;
	//----------------------------------------------------------------
	//	Underlying size
	//----------------------------------------------------------------
	protected int			m_size;

	//----------------------------------------------------------------
	//	Incomplete, either extern or in ca se of structs being currently defined
	//----------------------------------------------------------------
	protected boolean		m_incomplete = false;


	//----------------------------------------------------------------
	//
	//----------------------------------------------------------------
	public 
	Type ()
	{
		this("type");
	}

	public 
	Type (String strName)
	{
		this(strName, 0);
	}


	public 
	Type (String strName, int size)
	{
		setName(strName);
		setSize(size);
	}


	//----------------------------------------------------------------
	//
	//----------------------------------------------------------------
	public String
	getName ()
	{
		return m_typeName;
	}

	protected void
	setName (String str)
	{
		m_typeName = str;
	}


	//----------------------------------------------------------------
	//
	//----------------------------------------------------------------
	public int
	getSize ()
	{
		return m_size;
	}

	protected void
	setSize (int size)
	{
		m_size = size;
	}

	//----------------------------------------------------------------
	//
	//----------------------------------------------------------------
	public boolean
	isIncomplete ()
	{
		return m_incomplete;
	}

	public void
	setIsIncomplete (boolean inc)
	{
		m_incomplete = inc;
	}

	//----------------------------------------------------------------
	//
	//----------------------------------------------------------------
	public String
	toString()
	{
		return m_typeName;
	}

	//gets the mangled name for function overloading, just return type info in this case
	public String
	getMangledName()
	{
		return m_typeName;
	}

	//----------------------------------------------------------------
	// define own getClass method, since getClass in Object is final
	// and thus can't be overriden in AliasType
	//----------------------------------------------------------------
	public Class
	getClazz(){
		return this.getClass();
	}

	//gets the aliased type if this is an alias
	public Type
	getRootType(){
		return this;
	}
	//gets the pointed to type if this is a pointer
	public Type
	getPointedToType ()
	{
		return	this;
	}
	//get underlying type if this is array
	public Type
	getUnderlyingType ()
	{
		return this;
	}
	//get return type if this is func proto
	public Type
	getReturnType ()
	{
		return this;
	}

	//in case this is pointer this will be overriden, or alias as well
	public int getDerefs () { return 0; }
	public int getRootDerefs () { return getDerefs(); }

	//----------------------------------------------------------------
	//
	//----------------------------------------------------------------
	public boolean isEquivalent(Type t){
		if (t == null) return false;
		return (this.getClazz() == t.getClazz());
	}

	public boolean isAssignable(Type t){
		return isEquivalent(t);
	}	

	//types should be singletons anyway
	public Object 
	clone(){
		return this;
	}
	
	//----------------------------------------------------------------

	public boolean
	isCastable(Type t){
		return false;
	}

/*	
	//the srcSTO will be a const of this type's type
	//target type is the type that we want to cast to
	public ConstSTO
	castFrom (ConstSTO src, Type targetType){
		return src;
	}

	//the srcSTO will be a const of any castable type
	//target type will be this
	public ConstSTO
	castTo (ConstSTO src){
		return src;
	}
*/	

	//----------------------------------------------------------------
	//	It will be helpful to ask a Type what specific Type it is.
	//	The Java operator instanceof will do this, but you may
	//	also want to implement methods like isNumeric(), isInt(),
	//	etc. Below is an example of isInt(). Feel free to
	//	change this around.
	//----------------------------------------------------------------
	//first level hierarchy
	public boolean	isBasic ()		{ return false; }
	public boolean	isVoid ()		{ return false; }
//	public boolean	isNull ()		{ return false; }
	public boolean	isComposite ()	{ return false; }
	public boolean	isError ()		{ return false; }
	public boolean	isAlias ()		{ return false; }
	public boolean	isFunction ()	{ return false; }
	
	//second level hierarchy
	//basic
	public boolean	isNumeric ()	{ return false; }
	public boolean	isBool ()		{ return false; }
	public boolean	isChar ()		{ return false; }

	//comp
	public boolean	isArray ()		{ return false; }
	public boolean	isStruct ()		{ return false; }
	public boolean	isPtr ()		{ return false; }
	public boolean	isString ()		{ return false; }

	//third level hierarchy
	//numeric
	public boolean	isInt ()		{ return false; }
	public boolean	isFloat ()		{ return false; }

	//pointer
	public boolean	isBasicPtr ()	{ return false; }
	public boolean	isFuncPtr ()	{ return false; }
	public boolean	isJoker ()		{ return false; }


}
